from mhcgnomes import Allele, Gene, Species, parse

from .common import eq_


def test_macaque_allele_B_08_02():
    allele_name = "Mamu-B*082:02"
    eq_(parse(allele_name).to_string(), "Mamu-B*082:02")
    eq_(parse(allele_name).compact_string(), "B08202")


def test_macaque_allele_B_007_02():
    # expect 3rd zero in the family "007" to be trimmed in the normalized form
    # of this allele
    allele_name = "Mamu-B*007:02"
    eq_(parse(allele_name).to_string(), "Mamu-B*007:02")
    eq_(parse(allele_name).compact_string(), "B00702")


def test_gelada_species_Thge():
    expected = Species.get("Thge")
    assert expected is not None
    eq_(parse("Thge"), expected)


def test_gelada_gene_Thge_DQA1():
    expected = Gene.get("Thge", "DQA1")
    assert expected is not None
    eq_(parse("Thge-DQA1"), expected)


def test_gelada_allele_Thge_DQA1_25_01():
    expected = Allele.get("Thge", "DQA1", "25", "01")
    assert expected is not None
    eq_(parse("Thge-DQA1*25:01"), expected)


def test_monkey_species_Aoni():
    expected = Species.get("Aoni")
    assert expected is not None
    eq_(parse("Aoni"), expected)


def test_monkey_gene_Aoni_DRB3():
    expected = Gene.get("Aoni", "DRB3")
    assert expected is not None
    eq_(parse("Aoni-DRB3"), expected)


def test_monkey_allele_Aoni_DRB3_06_01():
    expected = Allele.get("Aoni", "DRB3", "06", "01")
    assert expected is not None
    eq_(parse("Aoni-DRB3*06:01"), expected)


def test_Mamu_A7_allele_without_seps():
    s = "Mamu-A70103"
    eq_(parse(s).to_string(), "Mamu-A7*01:03")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from collections import OrderedDict, defaultdict
from collections.abc import Iterable, Mapping
from typing import Any, Union

from .common import cache
from .data import allele_aliases as raw_allele_aliases_dict
from .data import gene_aliases as raw_gene_aliases_dict
from .data import haplotypes as raw_haplotypes_dict
from .data import heterodimers as raw_heterodimers_dict
from .data import known_alleles as raw_known_alleles_dict
from .data import serotypes as raw_serotypes_dict
from .data import species as raw_species_dict
from .data import supertypes as raw_supertypes_dict
from .mhc_class_helpers import class1_restrictions, class2_restrictions
from .normalizing_dictionary import NormalizingDictionary
from .normalizing_set import NormalizingSet
from .result import Result


class Species(Result):
    """
    Representation of a parsed species prefix such as "HLA", "ELA"
    """

    def __init__(
        self,
        name: str,
        common_name: str,
        mhc_prefix: str,
        gene_names: Iterable[str],
        gene_name_to_mhc_class: Mapping[str, str],
        class2_loci: Iterable[str],
        class2_locus_to_gene_names: Mapping[str, Iterable[str]],
        class2_gene_name_to_chain_type: Mapping[str, str],
        gene_aliases: Mapping[str, str],
        allele_aliases: Mapping[str, str],
        known_alleles: Mapping[str, Iterable[str]],
        haplotypes: Mapping[str, Iterable[str]],
        serotypes: Mapping[str, Iterable[str]],
        heterodimers: Mapping[str, Mapping[str, str]],
        supertypes: Mapping[str, Mapping[str, Any]],
        parent_species: Union["Species", None] = None,
        old_mhc_prefix: Union[str, None] = None,
        other_mhc_prefixes: Iterable[str] = [],
        other_common_names: Iterable[str] = [],
        raw_string: Union[str, None] = None,
    ):
        Result.__init__(self, raw_string=raw_string)
        self.name = name
        self.common_name = common_name
        self.other_common_names = list(other_common_names)
        self.mhc_prefix = mhc_prefix
        self.other_mhc_prefixes = set(other_mhc_prefixes)
        if old_mhc_prefix:
            self.old_mhc_prefix = old_mhc_prefix
        else:
            self.old_mhc_prefix = mhc_prefix

        self.gene_names = gene_names
        self.gene_name_to_mhc_class = gene_name_to_mhc_class
        self.class2_loci = class2_loci
        self.class2_locus_to_gene_names = class2_locus_to_gene_names
        self.class2_gene_name_to_chain_type = class2_gene_name_to_chain_type
        self.gene_aliases = gene_aliases
        # create a reverse lookup from proper names to their list of aliases
        self.reverse_gene_aliases = self._create_reverse_gene_aliases(gene_names, gene_aliases)

        self.allele_aliases = allele_aliases
        self.known_alleles = known_alleles
        self.haplotypes = haplotypes
        self.serotypes = serotypes
        self.heterodimers = heterodimers
        self.supertypes = supertypes
        self.parent_species = parent_species

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if type(other) is not Species:
            return False
        return self.name == other.name

    @classmethod
    def _create_reverse_gene_aliases(cls, gene_names, gene_aliases):
        """
        Creates dictionary mapping canonical gene name to its set of aliases.
        """
        d = defaultdict(NormalizingSet)

        for k, v in gene_aliases.items():
            d[v].add(k)
        for v in gene_names:
            d[v].add(v)
        return d

    @classmethod
    def str_field_names(cls):
        return ("name", "mhc_prefix")

    @classmethod
    def tuple_field_names(cls):
        return ("name", "mhc_prefix")

    @classmethod
    def eq_field_names(cls):
        return ("name",)

    @classmethod
    def from_dict(cls, d):
        if "name" in d:
            species_name = d["name"]
            return species_name_to_species_object.get(species_name)
        return None

    @property
    def species_name(self):
        return self.name

    @property
    def canonical_mhc_prefix(self):
        return self.mhc_prefix

    @property
    def parent(self):
        return self.parent_species

    @property
    def historic_mhc_prefix(self):
        """
        Return older species name which is now used to group multiple
        related species (e.g. "DLA" for "Calu").
        """
        return self.old_mhc_prefix

    @property
    def historic_alias(self):
        return self.historic_mhc_prefix

    @property
    def gene_names_and_aliases(self):
        name_set = set(self.gene_aliases.keys()).union(self.gene_names)
        return sorted(name_set)

    @property
    def num_genes(self):
        return len(self.gene_names)

    @property
    def scientific_species_name(self):
        return self.name

    @property
    def latin_name(self):
        return self.name

    @property
    def species_prefix(self):
        return self.canonical_mhc_prefix

    @property
    def prefix(self):
        return self.species_prefix

    def to_string(self, include_species=True, use_old_species_prefix=False):
        if not include_species:
            return ""
        elif use_old_species_prefix:
            return self.historic_alias
        else:
            return self.prefix

    def compact_string(self, include_species=False, use_old_species_prefix=False):
        return self.to_string(
            include_species=include_species, use_old_species_prefix=use_old_species_prefix
        )

    @classmethod
    @cache
    def get_multiple(cls, species_name):
        """
        Cached wrapper around find_matching_species which tries to
        identify one or more species from any of its scientific name,
        common names, MHC prefixes.
        """
        return find_matching_species_objects(species_name)

    @classmethod
    @cache
    def get(cls, species_name):
        if type(species_name) is Species:
            return species_name
        elif species_name is None or type(species_name) is not str:
            return None

        species_objects = cls.get_multiple(species_name)
        if len(species_objects) == 0:
            return None
        # sort species by how well their prefix or name matches the given query
        # and how many genes they have a species with more curated genes
        # is more likely to be the one we're looking for in case of an
        # ambiguous MHC prefix
        return sorted(species_objects, key=create_species_sort_key(species_name))[-1]

    def to_record(self):
        return OrderedDict(
            [
                ("species_prefix", self.prefix),
                ("species_name", self.name),
            ]
        )

    @property
    def common_species_name(self):
        """
        Returns common species name associated with MHC species
        prefix.
        """
        return self.common_name

    @property
    def all_mhc_prefixes(self):
        """
        Returns all MHC prefixes used for this species
        """
        prefixes = set(self.other_mhc_prefixes)
        if self.old_mhc_prefix:
            prefixes.add(self.old_mhc_prefix)
        return [self.prefix, *sorted(prefixes)]

    @property
    def all_common_names(self):
        """
        Returns all common names used for this species
        """
        return [self.common_name, *self.other_common_names]

    @property
    def all_identifiers(self):
        """
        Return all names and prefixes associated with this Species.

        Returns list of str
        """
        return [self.name, *self.all_mhc_prefixes, *self.all_common_names]

    def find_matching_gene_name(self, gene_name):
        """
        Use known aliases and normalized capitalization to infer
        the canonical gene name corresponding to the input.

        Returns str or None
        """
        if type(gene_name) in (int, float):
            gene_name = str(gene_name)
        if gene_name in self.gene_names:
            return self.gene_names.get_original(gene_name)
        elif gene_name in self.gene_aliases:
            alt_gene_name = self.gene_aliases[gene_name]
            if alt_gene_name in self.gene_names:
                # make sure the new gene name is a gene name
                # and not a locus
                # e.g. H2-IA -> H2-A is a mapping of loci
                return alt_gene_name
        return None

    def find_matching_class2_locus_name(self, locus_name):
        """
        Use known aliases and normalized capitalization to infer
        the canonical gene name corresponding to the input.

        Returns str or None
        """
        if type(locus_name) in (int, float):
            locus_name = str(locus_name)
        if locus_name in self.class2_loci:
            return self.class2_loci.get_original(locus_name)
        elif locus_name in self.gene_aliases:
            alt_locus_name = self.gene_aliases[locus_name]
            if alt_locus_name in self.class2_loci:
                # make sure the new locus name is an actual locus
                # and not a gene
                # e.g.
                # H2-IA -> H2-A is a mapping of loci
                # but
                # H2-IAb -> H2-Ab is a mppaing of genes
                return alt_locus_name
        return None

    def normalize_gene_name_if_exists(self, gene_name):
        normalized_name = self.find_matching_gene_name(gene_name)

        if normalized_name:
            return normalized_name
        else:
            return gene_name

    def get_mhc_class_of_gene(self, gene_name):
        """
        Parameters
        ----------
        gene_name : str

        Returns either one of:
            {"I", "Ia", "Ib", "Ic", "Id", "II", IIa", "IIb", "other"}
        or None if gene can't be found
        """
        gene_name = self.normalize_gene_name_if_exists(gene_name)
        return self.gene_name_to_mhc_class.get(gene_name)

    def get_known_allele(self, gene_name, allele_name):
        gene_name_candidates = {gene_name}

        if gene_name is not None:
            gene_aliases = self.reverse_gene_aliases.get(gene_name, [])
            gene_name_candidates.update(gene_aliases)

        for gene_name in gene_name_candidates:
            known_alleles_for_gene_name = self.known_alleles.get(gene_name)
            if known_alleles_for_gene_name and allele_name in known_alleles_for_gene_name:
                known_allele_name = known_alleles_for_gene_name.get_original(allele_name)
                return (gene_name, known_allele_name)

        # if allele isn't in known_alleles but it's an alias for an allele which
        # is, then also return it
        for gene_name in gene_name_candidates:
            if gene_name:
                key = f"{gene_name}*{allele_name}"
            else:
                key = allele_name
            if key in self.allele_aliases:
                new_name = self.allele_aliases[key]
                for gene_name2 in gene_name_candidates:
                    if new_name in self.known_alleles.get(gene_name2, []):
                        known_allele_name = self.allele_aliases.original_key(key)
                        return (gene_name, known_allele_name)
        return None

    @cache
    def get_allele_alias(self, gene_name: str, allele_name: str):
        """
        Returns None if no alias found, otherwise pair of
        (gene name, allele name)
        """
        gene_name_candidates = {gene_name}

        if gene_name is not None:
            gene_aliases = self.reverse_gene_aliases.get(gene_name, [])
            gene_name_candidates.update(gene_aliases)

        # if allele isn't in known_alleles try looking at the keys of
        # allele_aliases
        for gene_name in gene_name_candidates:
            if gene_name:
                key = f"{gene_name}*{allele_name}"
            else:
                key = allele_name
            if key in self.allele_aliases:
                value = self.allele_aliases[key]
                if "*" in value:
                    parts = value.split("*")
                    if len(parts) != 2:
                        continue
                    return tuple(parts)
                else:
                    return (None, value)
        return None

    @classmethod
    def get_species_with_gene_name(self, gene_name):
        """
        Returns list of Species which have the given gene name
        """
        if gene_name in gene_name_to_species_objects:
            return list(gene_name_to_species_objects[gene_name])
        else:
            return []

    @property
    def is_mouse(self):
        return self.prefix in {"H-2", "H2"}

    @property
    def is_chicken(self):
        return self.prefix == "Gaga"

    @property
    def is_rat(self):
        return self.prefix in {"RT1", "Rano"}

    @property
    def is_human(self):
        return self.prefix == "HLA"

    @property
    def is_dog(self):
        return self.prefix in {"DLA", "Calu"}

    @property
    def is_cat(self):
        return self.prefix in {"FLA", "Feca"}

    @property
    def is_pig(self):
        return self.prefix in {"SLA", "Susc"}

    @property
    def is_cow(self):
        return self.prefix == "BoLA"

    @property
    def is_horse(self):
        return self.prefix in {"ELA", "Eqca"}


################################################################################
#
# Parse the species/gene ontology and create dictionaries mapping
# to Species objects
#
################################################################################


def guess_class2_chain_type(gene_name):
    # For now we're guessing based on the name, e.g.
    # DRB1 is a beta chain and DRA is an alpha chain
    trimmed_gene_name = gene_name
    if trimmed_gene_name.endswith("like"):
        trimmed_gene_name = trimmed_gene_name[:-4]
    while trimmed_gene_name and trimmed_gene_name[-1].isdigit():
        trimmed_gene_name = trimmed_gene_name[:-1]
    if not trimmed_gene_name:
        return "beta"
    last_letter = trimmed_gene_name[-1].upper()
    is_alpha = last_letter == "A"
    # assume that anything which can't be pinned down to be an alpha chain
    # is then a beta chain, since more of the variability/gene copying
    # seems to occur in beta chains
    return "alpha" if is_alpha else "beta"


@cache
def create_species_for_latin_name(latin_name):
    if latin_name not in raw_species_dict:
        raise ValueError(f"Species not found: '{latin_name}'")
    species_info = raw_species_dict[latin_name]
    parent_species_latin_name = species_info.get("parent")
    if parent_species_latin_name:
        parent_species = create_species_for_latin_name(parent_species_latin_name)
    else:
        parent_species = None

    prefix = species_info.get("prefix")
    if not prefix:
        raise ValueError(f"Missing 'prefix' for '{latin_name}' in species ontology")

    old_mhc_prefix = species_info.get("old prefix")
    if not old_mhc_prefix and parent_species:
        old_mhc_prefix = parent_species.prefix

    other_mhc_prefixes = species_info.get("other prefixes")
    if type(other_mhc_prefixes) is str:
        other_mhc_prefixes = [other_mhc_prefixes]
    elif other_mhc_prefixes is None:
        other_mhc_prefixes = []

    common_name = species_info.get("name")
    if not common_name:
        raise ValueError(f"Missing 'name' for '{latin_name}' in species ontology")

    if type(common_name) is str:
        common_names = [common_name]
    else:
        common_names = common_name
    # make all common names lowercase
    common_names = [s.lower() for s in common_names]
    shortest_common_name = min(common_names, key=len)

    if parent_species is None:
        gene_names = NormalizingSet()
        gene_name_to_mhc_class = NormalizingDictionary()
        class2_loci = NormalizingSet()
        class2_locus_to_gene_names = NormalizingDictionary(default_value_fn=set)
        class2_gene_name_to_chain_type = NormalizingDictionary()
    else:
        gene_names = parent_species.gene_names.copy()
        gene_name_to_mhc_class = parent_species.gene_name_to_mhc_class.copy()
        class2_loci = parent_species.class2_loci.copy()
        class2_locus_to_gene_names = parent_species.class2_locus_to_gene_names.copy()
        class2_gene_name_to_chain_type = parent_species.class2_gene_name_to_chain_type.copy()

    for mhc_class, mhc_class_members in species_info.get("genes", {}).items():
        if mhc_class_members is None:
            raise ValueError(
                f"Unexpected None in gene ontology for class '{mhc_class}' of '{latin_name}'"
            )
        if mhc_class in class1_restrictions.union({"other"}):
            if type(mhc_class_members) is not list:
                raise ValueError(
                    f"Malformed gene ontology for '{latin_name}' MHC class '{mhc_class}'"
                )
            for gene_name in mhc_class_members:
                gene_names.add(str(gene_name))
                gene_name_to_mhc_class[gene_name] = mhc_class
        elif mhc_class in class2_restrictions:
            if type(mhc_class_members) is not dict:
                raise ValueError(
                    f"Malformed gene ontology for '{latin_name}' MHC class '{mhc_class}'"
                )
            for locus, locus_gene_names in mhc_class_members.items():
                class2_loci.add(locus)
                for gene_name in locus_gene_names:
                    gene_name = str(gene_name)
                    gene_names.add(gene_name)
                    gene_name_to_mhc_class[gene_name] = mhc_class
                    class2_locus_to_gene_names[locus].add(gene_name)
                    # TODO:
                    #  make alpha vs. beta chain genes explicit in
                    #  the YAML file ontology
                    chain_type = guess_class2_chain_type(gene_name)
                    class2_gene_name_to_chain_type[gene_name] = chain_type

    if parent_species is None:
        all_identifiers = []
    else:
        all_identifiers = list(parent_species.all_identifiers)
    all_identifiers.append(latin_name)
    all_identifiers.append(prefix)
    all_identifiers.extend(other_mhc_prefixes)
    all_identifiers.extend(common_names)

    # Don't rely on other YAML files to use any particular prefix or
    # species identifier, just try them all
    # One reason for this is that canonical species prefixes might change
    # over time and don't want one update to break code mysteriously
    # elsewhere. Another reason is that a few species share the same
    # prefix ('Bubu' belongs to both an owl species and water buffalo).
    gene_aliases = combine_species_aliases(raw_gene_aliases_dict, all_identifiers)
    allele_aliases = combine_species_aliases(raw_allele_aliases_dict, all_identifiers)
    haplotypes = combine_species_aliases(raw_haplotypes_dict, all_identifiers)
    serotypes = combine_species_aliases(raw_serotypes_dict, all_identifiers)
    heterodimers = combine_species_aliases(raw_heterodimers_dict, all_identifiers)
    supertypes = combine_species_aliases(raw_supertypes_dict, all_identifiers)
    known_alleles = combine_species_aliases(
        raw_known_alleles_dict, all_identifiers, value_class=NormalizingSet
    )

    return Species(
        name=latin_name,
        common_name=shortest_common_name,
        mhc_prefix=prefix,
        old_mhc_prefix=old_mhc_prefix,
        gene_names=gene_names,
        gene_name_to_mhc_class=gene_name_to_mhc_class,
        class2_loci=class2_loci,
        class2_locus_to_gene_names=class2_locus_to_gene_names,
        class2_gene_name_to_chain_type=class2_gene_name_to_chain_type,
        gene_aliases=gene_aliases,
        allele_aliases=allele_aliases,
        known_alleles=known_alleles,
        haplotypes=haplotypes,
        serotypes=serotypes,
        heterodimers=heterodimers,
        supertypes=supertypes,
        other_mhc_prefixes=other_mhc_prefixes,
        other_common_names=[name for name in common_names if name != shortest_common_name],
        raw_string=latin_name,
    )


def combine_species_aliases(
    species_dict, species_names, dictionary_class=NormalizingDictionary, value_class=None
):
    if value_class:
        result = dictionary_class(default_value_fn=value_class)
    else:
        result = dictionary_class()
    for species_name in species_names:
        value_for_species = species_dict.get(species_name)
        if value_for_species:
            if type(value_for_species) not in (dict, NormalizingDictionary):
                raise TypeError(f"Expected sub-dictionaries but got {type(value_for_species)}")
            for key, value in value_for_species.items():
                old_value = None
                if key in result:
                    old_value = result[key]
                elif value_class is not None:
                    old_value = value_class()

                if old_value is not None:
                    t = type(old_value)
                    if t in {set, dict, NormalizingSet, NormalizingDictionary}:
                        combined = old_value.copy()
                        combined.update(value)
                        value = combined
                    elif t in {list, tuple}:
                        value = old_value + t(value)
                result[key] = value
    return result


def create_species_lookup_dictionaries():
    gene_name_to_species_objects = NormalizingDictionary(default_value_fn=set)
    species_name_to_species_object = NormalizingDictionary()

    # latin name, common names, or MHC prefixes all mapping to multiple
    # species objects
    alias_to_species_objects = NormalizingDictionary(default_value_fn=set)

    for latin_name in raw_species_dict:
        species = create_species_for_latin_name(latin_name)
        species_name_to_species_object[latin_name] = species
        for s in species.all_identifiers:
            alias_to_species_objects[s].add(species)
        for gene_name in species.gene_names_and_aliases:
            gene_name_to_species_objects[gene_name].add(species)
    return (species_name_to_species_object, alias_to_species_objects, gene_name_to_species_objects)


species_name_to_species_object, alias_to_species_objects, gene_name_to_species_objects = (
    create_species_lookup_dictionaries()
)


def find_matching_species_objects(name):
    """
    Returns list of Species
    """
    if type(name) is Species:
        return [name]
    if name is None:
        return []
    return list(alias_to_species_objects.get(name, []))


def create_species_sort_key(query_string):
    """
    If we get multiple Species matching a query then
    use this key function to sort possible matches
    and prefer ones that match the query best.
    """

    def sort_key(species):
        same_prefix = species.prefix == query_string
        similar_prefix = species.prefix.lower().strip() == query_string.lower().strip()
        num_genes = species.num_genes
        return (same_prefix, similar_prefix, num_genes)

    return sort_key


def infer_species_from_prefix(name):
    """
    Trying to parse prefixes of alleles such as:
        HLA-A
    but also ones with dashes in the species prefix:
        H-2-K
    and also those lacking any dashes such as:
        H2K

     ...we also need to consider that alleles, haplotypes, etc may come
     immediately after the gene:
        H2Kk
        HLA-A0201

    Returns the Species and the original string that matched it or None.
    """
    # Try parsing a few different substrings to get the species,
    # and then use the species gene list to determine what the gene is in this string
    candidate_species_substrings = {name}

    if "-" in name or "." in name:
        # if name is "H-2-K" or "RT1.A" then try parsing "H", "H-2", "RT1" as
        # species prefixes. We treat both "-" and "." as valid separators.
        parts_split_by_separator = re.split(r"[-.]", name)
        candidate_species_substrings.add(parts_split_by_separator[0])
        if len(parts_split_by_separator) > 1:
            # Also try first two parts joined by hyphen (e.g., "H-2" from "H-2-K")
            candidate_species_substrings.add(
                parts_split_by_separator[0] + "-" + parts_split_by_separator[1]
            )

    for num_chars in [None, 4, 3, 2]:
        for candidate in candidate_species_substrings:
            if num_chars:
                original_prefix = candidate[:num_chars]
            else:
                original_prefix = candidate
            species_objects = find_matching_species_objects(original_prefix)
            if species_objects:
                if len(species_objects) > 1:
                    species_object = sorted(
                        species_objects, key=create_species_sort_key(original_prefix)
                    )[-1]
                else:
                    species_object = species_objects[0]
                return species_object, original_prefix

    # if all else fails, look for a distinctive gene name which is unique
    # to one species
    if "*" in name:
        parts = name.split("*")
        prefix = parts[0]
        species_objects = Species.get_species_with_gene_name(prefix)
        if len(species_objects) == 1:
            species = species_objects[0]
            # returning an empty string as the second result since no
            # part of the original string actually matched a species name
            # or alias but we inferred it from a gene name
            return species, ""
    return None



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Union

from .errors import ParseError
from .result import Result


class Mutation(Result):
    """
    Represents a single amino acid mutation in an MHC sequence.

    Mutations are described using standard notation: OriginalAA + Position + MutantAA.
    For example, "N80I" represents asparagine (N) at position 80 mutated to
    isoleucine (I).

    Parameters
    ----------
    pos : int
        The position of the mutation (1-indexed).
    aa_original : str
        The original amino acid (single letter code).
    aa_mutant : str
        The mutant amino acid (single letter code).
    raw_string : str, optional
        The original unparsed string.

    Attributes
    ----------
    pos : int
        The position of the mutation.
    aa_original : str
        The original amino acid.
    aa_mutant : str
        The mutant amino acid.

    Examples
    --------
    >>> mut = Mutation.parse("N80I")
    >>> mut.pos
    80
    >>> mut.aa_original
    'N'
    >>> mut.aa_mutant
    'I'
    >>> mut.to_string()
    'N80I'
    """

    def __init__(
        self, pos: int, aa_original: str, aa_mutant: str, raw_string: Union[str, None] = None
    ):
        Result.__init__(self, raw_string=raw_string)
        self.pos = pos
        self.aa_original = aa_original
        self.aa_mutant = aa_mutant

    def fields(self):
        return ("pos", "aa_original", "aa_mutant")

    def to_string(self):
        return f"{self.aa_original}{self.pos}{self.aa_mutant}"

    mutation_regex = re.compile(r"([a-yA-Y])(\d+)([a-yA-Y])")

    @classmethod
    def get(cls, pos, aa_original, aa_mutant, raw_string=None):
        pos = int(pos)
        aa_original = aa_original.upper()
        aa_mutant = aa_mutant.upper()
        return Mutation(pos, aa_original, aa_mutant, raw_string=raw_string)

    @classmethod
    def parse(cls, seq, raise_on_error=True):
        seq = seq.upper()
        result = cls.mutation_regex.fullmatch(seq)
        if result is None:
            if raise_on_error:
                raise ParseError(f"Allele mutation malformed: '{seq}'")
            else:
                return None
        return cls.get(
            pos=result.group(2),
            aa_original=result.group(1),
            aa_mutant=result.group(3),
            raw_string=seq,
        )



from mhcgnomes import Allele, Species, parse

from .common import eq_


def test_parse_DLA_as_dog_species():
    eq_(parse("DLA"), Species.get("DLA"))


def test_parse_Calu_as_dog_species():
    eq_(parse("Calu"), Species.get("Calu"))


def test_parse_dog_as_species():
    eq_(parse("dog"), Species.get("Calu"))


def test_parse_dog_class2_allele_dla_dqa1_001_01():
    eq_(parse("DLA-DQA1*00101"), Allele.get("DLA", "DQA1", "001", "01"))


def test_normalized_string_dog_class2_allele_dla_dqa1_001_01_yes_alias():
    eq_(parse("DLA-DQA1*00101").to_string(use_old_species_prefix=True), "DLA-DQA1*001:01")


def test_only_2_digits_in_first_allele_field():
    # make sure it parses at all
    eq_(parse("DLA-DQA1*0101"), Allele.get("DLA", "DQA1", "01", "01"))
    # TODO:
    #   once we have normalization of allele field lengths, check that
    #   first allele field becomes three digit
    #
    #   eq_(parse("DLA-DQA1*0101").allele_fields[0], "001")


def test_species_code_calu_no_alias():
    eq_(parse("Calu-DQA1*00101").to_string(use_old_species_prefix=False), "Calu-DQA1*001:01")


def test_species_code_calu_yes_alias():
    eq_(parse("Calu-DQA1*00101").to_string(use_old_species_prefix=True), "DLA-DQA1*001:01")


def test_parse_dog_class1_allele_dla_88_508_01():
    expected = Allele.get("DLA", "88", "508", "01")
    parsed = parse("DLA-88*50801")
    eq_(parsed, expected)


def test_compact_string_dog_class1_allele_dla_88_508_01():
    eq_(parse("DLA-88*50801").compact_string(), "88*50801")


def test_normalized_string_dog_class1_allele_dla_88_508_01_yes_alias():
    eq_(parse("DLA-88*50801").to_string(use_old_species_prefix=True), "DLA-88*508:01")


def test_normalized_string_dog_class1_allele_dla_88_508_01_no_alias():
    eq_(parse("DLA-88*50801").to_string(use_old_species_prefix=False), "DLA-88*508:01")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sequence
from typing import Union

from .allele import Allele
from .result_with_multiple_alleles import ResultWithMultipleAlleles
from .species import Species


class Serotype(ResultWithMultipleAlleles):
    """
    Represents an MHC serotype (serological antigen type).

    A Serotype represents a group of MHC alleles that are recognized by the
    same antibody. For example, "HLA-A2" is a serotype that includes multiple
    alleles like A*02:01, A*02:02, etc.

    Serotypes are commonly used in older literature and transplantation
    medicine before high-resolution molecular typing became available.

    Parameters
    ----------
    species : Species
        The species this serotype belongs to.
    name : str
        The serotype name (e.g., "A2", "B27", "DR4").
    alleles : Sequence[Allele]
        The alleles that belong to this serotype.
    raw_string : str, optional
        The original unparsed string.

    Attributes
    ----------
    species : Species
        The species this serotype belongs to.
    name : str
        The serotype name.
    alleles : Sequence[Allele]
        The alleles in this serotype.

    Examples
    --------
    >>> from mhcgnomes import parse
    >>> sero = parse("HLA-A2")
    >>> sero.name
    'A2'
    >>> len(sero.alleles) > 0
    True
    """

    def __init__(
        self,
        species: Species,
        name: str,
        alleles: Sequence[Allele],
        raw_string: Union[str, None] = None,
    ):
        ResultWithMultipleAlleles.__init__(
            self, species=species, name=name, alleles=alleles, raw_string=raw_string
        )

    def to_string(self, include_species=True, use_old_species_prefix=False):
        if include_species:
            return f"{self.species.to_string(use_old_species_prefix=use_old_species_prefix)}-{self.name}"
        else:
            return self.name

    def compact_string(self, include_species=False, use_old_species_prefix=False):
        return self.to_string(
            include_species=include_species, use_old_species_prefix=use_old_species_prefix
        )

    def to_record(self):
        d = self.species.to_record()
        d["serotype"] = self.to_string()
        return d



from mhcgnomes import Allele

from .common import eq_


def test_allele_get_A0201():
    allele = Allele.get("HLA", "A", "02", "01")
    assert allele is not None
    assert type(allele) is Allele
    eq_(allele.species_prefix, "HLA")
    eq_(allele.gene_name, "A")
    eq_(list(allele.allele_fields), ["02", "01"])
    eq_(allele.mhc_class, "Ia")


def test_restrict_num_allele_fields_A02010101():
    allele_eight_digit = Allele.get("HLA", "A", "02", "01", "01", "01")
    assert allele_eight_digit is not None
    assert type(allele_eight_digit) is Allele
    eq_(allele_eight_digit.num_allele_fields, 4)
    allele_four_digit = allele_eight_digit.restrict_allele_fields(2)
    eq_(allele_four_digit.num_allele_fields, 2)


def test_no_annotations():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01")
    assert not allele.annotation_null
    assert not allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert not allele.annotation_secreted
    assert not allele.annotation_pseudogene
    assert not allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert not allele.annotation_group
    assert not allele.annotation_splice_variant


def test_annotation_null():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01", annotation="N")
    assert allele.annotation_null
    assert not allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert not allele.annotation_secreted
    assert not allele.annotation_pseudogene
    assert not allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert not allele.annotation_group
    assert not allele.annotation_splice_variant


def test_annotation_cytosolic():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01", annotation="C")
    assert not allele.annotation_null
    assert allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert not allele.annotation_secreted
    assert not allele.annotation_pseudogene
    assert not allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert not allele.annotation_group
    assert not allele.annotation_splice_variant


def test_annotation_secreted():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01", annotation="S")
    assert not allele.annotation_null
    assert not allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert allele.annotation_secreted
    assert not allele.annotation_pseudogene
    assert not allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert not allele.annotation_group
    assert not allele.annotation_splice_variant


def test_annotation_questionable():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01", annotation="Q")
    assert not allele.annotation_null
    assert not allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert not allele.annotation_secreted
    assert not allele.annotation_pseudogene
    assert allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert not allele.annotation_group
    assert not allele.annotation_splice_variant


def test_annotation_group():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01", annotation="G")
    assert not allele.annotation_null
    assert not allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert not allele.annotation_secreted
    assert not allele.annotation_pseudogene
    assert not allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert allele.annotation_group
    assert not allele.annotation_splice_variant


def test_annotation_splice_variant():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01", annotation="Sp")
    assert not allele.annotation_null
    assert not allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert not allele.annotation_secreted
    assert not allele.annotation_pseudogene
    assert not allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert not allele.annotation_group
    assert allele.annotation_splice_variant


def test_annotation_pseudogene():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01", annotation="Ps")
    assert not allele.annotation_null
    assert not allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert not allele.annotation_secreted
    assert allele.annotation_pseudogene
    assert not allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert not allele.annotation_group
    assert not allele.annotation_splice_variant



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def strip_chars(s: str, chars_to_remove, _cache=None):
    if _cache is None:
        _cache = {}
    if s in _cache:
        return _cache[s]
    original = s
    for c in chars_to_remove:
        while s and s[0] == c:
            s = s[1:]
        while s and s[-1] == c:
            s = s[:-1]
    _cache[original] = s
    return s


def strip_whitespace_and_dashes(s: str):
    return strip_chars(s, "- ")


def smart_split(seq: str, sep: str):
    """
    Split string on a separator and then get rid of dashes
    and empty strings
    """
    return [strip_whitespace_and_dashes(p) for p in seq.split(sep)]


def split_on_all_seps(seq: str, seps="_:"):
    """
    Split given string on all separators specified

    For example, 02_01:01 will be split_token_sequences into:
        ["02", "01", "01"]
    """
    parts = [seq]
    for sep in seps:
        new_parts = []
        for subseq in parts:
            new_parts.extend(smart_split(subseq, sep))
        parts = new_parts
    return parts


def split_digits_at_end(seq: str):
    """
    Splits strings like "A0201" into ("A", "0201")
    """
    prefix = seq
    suffix = ""
    while prefix and prefix[-1].isdigit():
        suffix = prefix[-1] + suffix
        prefix = prefix[:-1]
    return prefix, suffix


def contains_any_letters(s: str):
    """
    Returns True if any characters in the sequence are letters.
    """
    return any(si.isalpha() for si in s)


def contains_whitespace(s: str):
    """
    Returns True if any whitespace chars in input string.
    """
    return " " in s or "\t" in s


def split_allele_fields(
    str_after_gene, allow_three_digits_in_first_field, allow_three_digits_in_second_field
):
    if ":" in str_after_gene:
        return str_after_gene.split(":")

    # if we don't have ':' to guide the field boundaries
    # then split_token_sequences on all possible seps and try to guess
    # which blocks of numbers are actually multiple fields.
    parts = split_on_all_seps(str_after_gene)

    parsed_fields = []
    failed = False
    if len(parts) == 1:
        part = parts[0]
        if len(part) in {2, 3}:
            return [part]

    for part in parts:
        if failed:
            break
        if part.isdigit():
            if allow_three_digits_in_first_field and len(parsed_fields) == 0 and len(part) > 4:
                parsed_fields.append(part[:3])
                part = part[3:]
            if allow_three_digits_in_second_field and len(parsed_fields) == 1 and len(part) > 4:
                parsed_fields.append(part[3:])

            while part and not failed:
                n_parsed = len(parsed_fields)
                remaining_length = len(part)
                if remaining_length == 1:
                    failed = True
                    break
                if (
                    allow_three_digits_in_first_field
                    and n_parsed == 0
                    and (remaining_length == 3 or remaining_length > 4)
                ) or (
                    allow_three_digits_in_second_field
                    and n_parsed == 1
                    and (remaining_length == 3 or remaining_length > 4)
                ):
                    boundary_index = 3
                else:
                    boundary_index = 2
                parsed_fields.append(part[:boundary_index])
                part = part[boundary_index:]
        else:
            parsed_fields.append(part)
    # if failed to parse anything then back up and try to turn some of
    # the optional arguments false
    if not failed and len(parsed_fields) > 0:
        return parsed_fields
    elif allow_three_digits_in_first_field:
        return split_allele_fields(
            str_after_gene,
            allow_three_digits_in_first_field=False,
            allow_three_digits_in_second_field=allow_three_digits_in_second_field,
        )
    elif allow_three_digits_in_second_field:
        return split_allele_fields(
            str_after_gene,
            allow_three_digits_in_first_field=False,
            allow_three_digits_in_second_field=False,
        )
    return None



"""
Tests for MHC nomenclature patterns found in scientific literature and IEDB.

This test file validates parsing of nomenclature patterns commonly encountered
in immunology publications, clinical reports, and databases like IEDB.

Sources and References:
-----------------------
HLA Nomenclature:
    - IPD-IMGT/HLA Database: https://www.ebi.ac.uk/ipd/imgt/hla/
    - HLA Nomenclature: https://hla.alleles.org/
    - History of HLA naming: https://en.wikipedia.org/wiki/History_and_naming_of_human_leukocyte_antigens

HLA Supertypes:
    - Sidney J, et al. "HLA class I supertypes: a revised and updated classification."
      BMC Immunol. 2008;9:1. PMC2245908
      https://pmc.ncbi.nlm.nih.gov/articles/PMC2245908/
    - Sette A, Sidney J. "Nine major HLA class I supertypes account for the vast
      preponderance of HLA-A and -B polymorphism." Immunogenetics. 1999;50:201-12.
      https://pubmed.ncbi.nlm.nih.gov/10602880/

KIR Ligand Groups (C1/C2, Bw4/Bw6):
    - IPD-KIR Ligand Calculator: https://www.ebi.ac.uk/ipd/kir/ligand.html
    - Gwozdowicz S, et al. "KIR specificity and avidity of standard and unusual
      C1, C2, Bw4, Bw6 and A3/11 amino acid motifs." Int J Immunogenet. 2019.
      https://pubmed.ncbi.nlm.nih.gov/31210416/

Non-Human MHC:
    - IPD-MHC Database: https://www.ebi.ac.uk/ipd/mhc/
    - Klein J, et al. "Nomenclature for the major histocompatibility complexes
      of different species." Immunogenetics. 1990;31:217-9.

Disease Associations:
    - HLA-B27 Syndromes: https://www.ncbi.nlm.nih.gov/books/NBK551523/
    - Celiac disease and DQ2/DQ8: PMC3482388

IEDB Data:
    - mhc_ligand_full.csv from https://www.iedb.org/
"""

import pytest

from mhcgnomes import parse
from mhcgnomes.allele import Allele
from mhcgnomes.mhc_class import MhcClass
from mhcgnomes.pair import Pair
from mhcgnomes.serotype import Serotype
from mhcgnomes.species import Species


class TestHumanSerotypesFromIEDB:
    """
    Test human HLA serotypes frequently found in IEDB data.

    IEDB occurrence counts (mhc_ligand_full.csv, top serotype patterns):
        HLA-B27:  211,287
        HLA-A2:   188,953
        HLA-A3:    71,107
        HLA-B35:   63,003
        HLA-A11:   58,725
        HLA-B51:   56,390
        HLA-B7:    53,047
        HLA-A1:    50,436
        HLA-B44:   47,475
        HLA-B15:   44,536
    """

    @pytest.mark.parametrize(
        "name,expected_serotype",
        [
            # Top 10 serotypes from IEDB
            ("HLA-B27", "B27"),
            ("HLA-A2", "A2"),
            ("HLA-A3", "A3"),
            ("HLA-B35", "B35"),
            ("HLA-A11", "A11"),
            ("HLA-B51", "B51"),
            ("HLA-B7", "B7"),
            ("HLA-A1", "A1"),
            ("HLA-B44", "B44"),
            ("HLA-B15", "B15"),
            # Without prefix
            ("B27", "B27"),
            ("A2", "A2"),
        ],
    )
    def test_top_serotypes_from_iedb(self, name, expected_serotype):
        """Top serotypes from IEDB should parse correctly."""
        result = parse(name)
        assert isinstance(result, Serotype)
        assert result.name == expected_serotype


class TestHumanAllelesFromIEDB:
    """
    Test human HLA alleles frequently found in IEDB data.

    IEDB occurrence counts (top alleles):
        HLA-A*02:01:  39,006
        HLA-B*07:02:  23,792
        HLA-A*01:01:  19,677
        HLA-B*27:05:  17,403
        HLA-A*24:02:  13,742
    """

    @pytest.mark.parametrize(
        "name",
        [
            "HLA-A*02:01",
            "HLA-B*07:02",
            "HLA-A*01:01",
            "HLA-B*27:05",
            "HLA-A*24:02",
            "HLA-B*57:01",
            "HLA-A*03:01",
            "HLA-A*11:01",
            # Without HLA prefix
            "A*02:01",
            "B*07:02",
        ],
    )
    def test_top_alleles_from_iedb(self, name):
        """Top alleles from IEDB should parse as Allele objects."""
        result = parse(name)
        assert isinstance(result, Allele)


class TestClassIIPairsFromIEDB:
    """
    Test Class II alpha/beta pairs from IEDB.

    IEDB occurrence counts:
        HLA-DPA1*01:03/DPB1*04:01:  9,237
        HLA-DPA1*02:02/DPB1*05:01:  6,020
        HLA-DRA*01:01/DRB1*01:01:   5,888
    """

    @pytest.mark.parametrize(
        "name",
        [
            "HLA-DPA1*01:03/DPB1*04:01",
            "HLA-DPA1*02:02/DPB1*05:01",
            "HLA-DRA*01:01/DRB1*01:01",
            "HLA-DQA1*01:02/DQB1*06:02",
            # Without HLA prefix
            "DPA1*01:03/DPB1*04:01",
            "DRA*01:01/DRB1*01:01",
        ],
    )
    def test_class2_pairs_from_iedb(self, name):
        """Class II pairs from IEDB should parse as Pair objects."""
        result = parse(name)
        assert isinstance(result, Pair)


class TestMhcClassFromIEDB:
    """
    Test MHC class nomenclature from IEDB.

    IEDB occurrence counts:
        HLA class I:   96,221
        HLA class II:  32,913
    """

    @pytest.mark.parametrize(
        "name,expected_class",
        [
            ("HLA class I", "I"),
            ("HLA class II", "II"),
            ("MHC class I", "I"),
            ("MHC class II", "II"),
            ("class I", "I"),
            ("class II", "II"),
        ],
    )
    def test_mhc_class_from_iedb(self, name, expected_class):
        """MHC class notation should parse as MhcClass."""
        result = parse(name)
        assert isinstance(result, MhcClass)
        assert result.mhc_class == expected_class


class TestMouseH2FromIEDB:
    """
    Test mouse H-2 nomenclature from IEDB.

    Mouse MHC uses haplotype letters (a, b, d, k, q, s, etc.).

    IEDB occurrence counts:
        H2-Db:   6,257 (class I, b haplotype)
        H2-Kb:   6,082 (class I, b haplotype)
        H2-Kk:   ~600  (class I, k haplotype)
        H2-IAb:  ~100  (class II, b haplotype)
        H2-IAg7: ~100  (class II, NOD mouse)
    """

    @pytest.mark.parametrize(
        "name",
        [
            # Class I
            "H2-Db",
            "H2-Kb",
            "H2-Kd",
            "H2-Kk",
            "H2-Ld",
            "H2-Dd",
            # Class II
            "H2-IAb",
            "H2-IAd",
            "H2-IAk",
            "H2-IEd",
            "H2-IEk",
            # Special: NOD mouse diabetes model
            "H2-IAg7",
            "H2-IEg7",
        ],
    )
    def test_mouse_h2_from_iedb(self, name):
        """Mouse H-2 nomenclature should parse correctly."""
        result = parse(name)
        # H2-D, H2-K, H2-L are class I (parse as Allele)
        # H2-IA, H2-IE are class II (parse as Pair or Haplotype)
        assert result is not None
        assert result.species.common_name.lower() == "mouse"


class TestNonHumanPrimatesFromIEDB:
    """
    Test non-human primate MHC nomenclature.

    Species prefixes:
    - Mamu: Macaca mulatta (rhesus macaque)
    - Patr: Pan troglodytes (chimpanzee)
    - Gogo: Gorilla gorilla (gorilla)

    IEDB occurrence counts (sample):
        Mamu-A1*001:01: 1,708
        Mamu-B*017:04:    717
        Mamu-B*17:        100

    References:
    - IPD-MHC NHP: https://www.ebi.ac.uk/ipd/mhc/group/NHP/
    """

    @pytest.mark.parametrize(
        "name",
        [
            # Rhesus macaque
            "Mamu-A1*001:01",
            "Mamu-A1*011:01",
            "Mamu-B*017:04",
            "Mamu-B*001:01",
            "Mamu-B*17",
            "Mamu-A*01",
            # Chimpanzee
            "Patr-A*01:01",
            "Patr-B*22:04",
        ],
    )
    def test_nhp_alleles_from_iedb(self, name):
        """Non-human primate MHC alleles should parse correctly."""
        result = parse(name)
        assert isinstance(result, Allele)


class TestKIRLigandGroups:
    """
    Test KIR ligand group nomenclature (Bw4, Bw6, C1, C2).

    Bw4 and Bw6 are public epitopes (supertypic specificities) on HLA-B.
    C1 and C2 are KIR ligand groups based on HLA-C polymorphism at position 80.

    References:
    - IPD-KIR Ligand Calculator: https://www.ebi.ac.uk/ipd/kir/ligand.html
    - Gwozdowicz 2019: https://pubmed.ncbi.nlm.nih.gov/31210416/
    """

    @pytest.mark.parametrize(
        "name",
        [
            # Public epitopes (parse as serotypes)
            "Bw4",
            "Bw6",
            "HLA-Bw4",
            "HLA-Bw6",
        ],
    )
    def test_bw4_bw6_public_epitopes(self, name):
        """Bw4 and Bw6 public epitopes should parse as serotypes."""
        result = parse(name)
        assert isinstance(result, Serotype)


class TestHLASupertypes:
    """
    Test HLA supertype nomenclature.

    Supertypes group HLA alleles with similar peptide binding properties.
    Nine major supertypes: A1, A2, A3, A24, B7, B27, B44, B58, B62.

    IEDB occurrence counts:
        B44 supertype: 966
        A2 supertype:  355
        A3 supertype:  221
        A1 supertype:  124
        B7 supertype:   64

    References:
    - Sidney 2008: https://pmc.ncbi.nlm.nih.gov/articles/PMC2245908/
    - Sette & Sidney 1999: https://pubmed.ncbi.nlm.nih.gov/10602880/
    """

    @pytest.mark.parametrize(
        "name",
        [
            "A2 supertype",
            "A3 supertype",
            "B7 supertype",
            "B44 supertype",
            "A1 supertype",
            "A24 supertype",
        ],
    )
    def test_hla_supertypes(self, name):
        """HLA supertype nomenclature should parse as Supertype objects."""
        from mhcgnomes import Supertype

        result = parse(name)
        assert result is not None
        assert isinstance(result, Supertype)


class TestMutantAlleles:
    """
    Test mutant allele nomenclature from IEDB.

    IEDB includes mutant alleles with notation like "HLA-B*27:05 C67S mutant".

    IEDB occurrence counts:
        HLA-B*27:05 C67S mutant:           1,797
        HLA-DRA*01:01/DRB1*01:01 C30S mutant: 888
    """

    @pytest.mark.parametrize(
        "name",
        [
            "HLA-B*27:05 C67S mutant",
            "HLA-A*02:01 Y84A mutant",
        ],
    )
    def test_mutant_alleles(self, name):
        """Mutant allele nomenclature should parse as Allele with mutations."""
        result = parse(name)
        assert isinstance(result, Allele)
        assert len(result.mutations) > 0


class TestWorkshopSerotypes:
    """
    Test workshop (w) serotype nomenclature.

    Workshop serotypes use "w" prefix to indicate provisional designations.
    Common in HLA-C (Cw) and HLA-DP (DPw).
    """

    @pytest.mark.parametrize(
        "name,expected_serotype",
        [
            ("HLA-Cw7", "Cw7"),
            ("HLA-Cw15", "Cw15"),
            ("Cw7", "Cw7"),
            ("HLA-DPw4", "DPw4"),
            ("HLA-Aw68", "Aw68"),
        ],
    )
    def test_workshop_serotypes(self, name, expected_serotype):
        """Workshop serotypes should parse correctly."""
        result = parse(name)
        assert isinstance(result, Serotype)
        assert result.name == expected_serotype


class TestSerotypeVariants:
    """
    Test serotype variants with letter suffixes.

    Some serotypes have letter suffixes indicating variants (e.g., B27c, DR3d).
    """

    @pytest.mark.parametrize(
        "name,expected_serotype",
        [
            ("HLA-B27c", "B27c"),
            ("HLA-DR3d", "DR3d"),
            ("B27c", "B27c"),
        ],
    )
    def test_serotype_variants(self, name, expected_serotype):
        """Serotype variants with letter suffixes should parse."""
        result = parse(name)
        assert isinstance(result, Serotype)
        assert result.name == expected_serotype


class TestNonClassicalHLA:
    """
    Test non-classical HLA molecules (HLA-E, HLA-F, HLA-G).

    Non-classical MHC-Ib molecules have limited polymorphism and
    specialized immune functions.

    References:
    - Evolution of MHC-G, -E, -F: PMC9352621
    """

    @pytest.mark.parametrize(
        "name",
        [
            "HLA-E*01:01",
            "HLA-F*01:01",
            "HLA-G*01:01",
            "HLA-G*01:04",
        ],
    )
    def test_non_classical_hla_alleles(self, name):
        """Non-classical HLA alleles should parse correctly."""
        result = parse(name)
        assert isinstance(result, Allele)


class TestDogDLA:
    """
    Test dog leukocyte antigen (DLA) nomenclature.

    IEDB occurrence:
        DLA-88*501:01: 1,640
    """

    @pytest.mark.parametrize(
        "name",
        [
            "DLA-88*501:01",
            "DLA-88*502:01",
        ],
    )
    def test_dog_dla_alleles(self, name):
        """Dog DLA alleles should parse correctly."""
        result = parse(name)
        assert isinstance(result, Allele)


class TestCompactAlleleNotation:
    """
    Test compact (no colon) allele notation.

    Older publications often use compact notation like A*0201 instead of A*02:01.
    """

    @pytest.mark.parametrize(
        "compact,standard",
        [
            ("A*0201", "A*02:01"),
            ("B*0702", "B*07:02"),
            ("DRB1*0401", "DRB1*04:01"),
            ("HLA-A*0201", "HLA-A*02:01"),
        ],
    )
    def test_compact_notation(self, compact, standard):
        """Compact notation should parse equivalently to standard notation."""
        compact_result = parse(compact)
        standard_result = parse(standard)

        assert isinstance(compact_result, Allele)
        assert isinstance(standard_result, Allele)
        # Should produce equivalent allele fields
        assert compact_result.allele_fields == standard_result.allele_fields


class TestDiseaseAssociatedAlleles:
    """
    Test disease-associated HLA alleles commonly cited in literature.

    References:
    - HLA-B27 and ankylosing spondylitis: PMID 4405240 (1973 discovery)
    - HLA-DQ2/DQ8 and celiac disease: PMC3482388
    - HLA-B*57:01 and abacavir hypersensitivity
    """

    @pytest.mark.parametrize(
        "name,disease",
        [
            ("HLA-B27", "ankylosing spondylitis"),
            ("HLA-B*27:05", "ankylosing spondylitis (major subtype)"),
            ("HLA-DQ2", "celiac disease"),
            ("HLA-DQ8", "celiac disease"),
            ("HLA-B*57:01", "abacavir hypersensitivity"),
            ("HLA-DR4", "rheumatoid arthritis"),
            ("HLA-DR3", "type 1 diabetes"),
        ],
    )
    def test_disease_associated_alleles(self, name, disease):
        """Disease-associated alleles commonly cited in literature should parse."""
        result = parse(name)
        assert result is not None, f"{name} ({disease}) should parse"


class TestSpeciesExist:
    """Test that key MHC species are defined."""

    @pytest.mark.parametrize(
        "prefix,common_name_contains",
        [
            ("HLA", "human"),
            ("H2", "mouse"),
            ("RT1", "rat"),
            ("Mamu", "rhesus"),
            ("Patr", "chimpanzee"),
            ("BoLA", "cattle"),
            ("SLA", "pig"),
            ("DLA", "canine"),
        ],
    )
    def test_species_defined(self, prefix, common_name_contains):
        """Key MHC species should be defined in mhcgnomes."""
        species = Species.get(prefix)
        assert species is not None, f"Species {prefix} should exist"
        assert common_name_contains in species.common_name.lower()



from mhcgnomes import Allele, parse

from .common import eq_


def test_sheep_class1_allele_vs_get():
    eq_(parse("Ovar-N*50001"), Allele.get("Ovar", "N", "500", "01"))


def test_sheep_class1_allele_string():
    eq_(parse("Ovar-N*50001").to_string(use_old_species_prefix=False), "Ovar-N*500:01")


def test_sheep_class1_allele_string_old_prefix():
    eq_(parse("Ovar-N*50001").to_string(use_old_species_prefix=True), "OLA-N*500:01")


def test_sheep_class2_allele_vs_get():
    eq_(parse("Ovar-DRB1*0804"), Allele.get("Ovar", "DRB1", "08", "04"))


def test_sheep_class2_allele_string():
    eq_(parse("Ovar-DRB1*0804").to_string(use_old_species_prefix=False), "Ovar-DRB1*08:04")


def test_sheep_class2_allele_string_old_prefix():
    eq_(parse("Ovar-DRB1*0804").to_string(use_old_species_prefix=True), "OLA-DRB1*08:04")



"""
Tests for mouse H2-Qa non-classical MHC genes.

The Qa genes (Qa-1, Qa-2, etc.) are non-classical MHC class Ib genes in mice.
They have official names (T23, Q7, etc.) but are commonly referred to by
their Qa names in the literature.

Gene aliases:
- Qa / Qa1 / Qa-1 -> T23
- Qa2 / Qa-2 -> Q7
- Qa10 / Qa-10 -> Q10
"""

import pytest

from mhcgnomes import Allele, Gene, parse


class TestMouseQaGenes:
    """Test parsing of mouse Qa gene nomenclature."""

    @pytest.mark.parametrize(
        "name,expected_gene",
        [
            ("H2-Qa", "T23"),
            ("H2-Qa1", "T23"),
            ("H2-Qa-1", "T23"),
            ("H2-Qa2", "Q7"),
            ("H2-Qa-2", "Q7"),
            ("H2-Qa10", "Q10"),
            ("H2-Qa-10", "Q10"),
        ],
    )
    def test_qa_genes(self, name, expected_gene):
        """Qa gene names should resolve to their official T/Q names."""
        result = parse(name)
        assert isinstance(result, Gene)
        assert result.name == expected_gene

    @pytest.mark.parametrize(
        "name,expected_gene,expected_allele",
        [
            # Valid notations for gene T23 (Qa-1), allele b
            ("H2-Qa-1b", "T23", ("b",)),
            ("H2-Qa1*b", "T23", ("b",)),
            ("H2-Qa*b", "T23", ("b",)),
            # Parenthesis notation also works
            ("H2-Qa-1(b)", "T23", ("b",)),
            ("H2-Qa1(b)", "T23", ("b",)),
        ],
    )
    def test_qa_alleles(self, name, expected_gene, expected_allele):
        """Qa allele names should parse to gene T23, allele b."""
        result = parse(name)
        assert isinstance(result, Allele)
        assert result.gene.name == expected_gene
        assert result.allele_fields == expected_allele

    def test_qa_star_1b_is_wrong_notation(self):
        """
        Qa*1b is wrong notation - the 1 is part of the gene name, not allele.

        This parses but gives the "wrong" result (allele "1b" instead of "b"),
        which is correct behavior - we shouldn't silently fix bad notation.
        The correct notation is Qa1*b or Qa-1b.
        """
        result = parse("H2-Qa*1b")
        assert isinstance(result, Allele)
        assert result.gene.name == "T23"
        # Note: allele is "1b" not "b" - this is expected for wrong notation
        assert result.allele_fields == ("1b",)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from collections.abc import Iterable
from typing import Optional

from .allele import Allele
from .allele_without_gene import AlleleWithoutGene
from .gene import Gene
from .haplotype import Haplotype
from .pair import Pair
from .result import Result
from .serotype import Serotype

# Pattern for serotype names that look like alleles (e.g., B3901, A2403, DR1403)
# These should have lower priority than actual allele interpretations
_ALLELE_LIKE_SEROTYPE_PATTERN = re.compile(r"^([ABC]\d{4,}|D[RPQO][AB]?\d{4,})$")


def pick_best_result(candidates: Iterable[Result], raise_on_error=True) -> Optional[Result]:
    if len(candidates) == 0:
        if raise_on_error:
            raise ValueError("Expected at least one candidate ParseResult object")
        else:
            return None
    sorted_candidates = sort_results(candidates)
    return sorted_candidates[0]


def sort_results(results: Iterable[Result]) -> list[Result]:
    if type(results) is not list:
        results = list(results)
    if not results or len(results) == 1:
        return results
    return sorted(results, key=sort_key, reverse=True)


def sort_key(result: Result):
    lower_raw_string = result.raw_string.lower() if result.raw_string else ""
    full_string_matches_raw_string = result.to_string().lower() == lower_raw_string
    compact_string_matches_raw_string = result.compact_string().lower() == lower_raw_string
    if hasattr(result, "name"):
        name_matches_raw_string = lower_raw_string == result.name.lower()
    else:
        name_matches_raw_string = False
    t = type(result)
    is_class2_pair = t is Pair
    alpha_is_allele = is_class2_pair and type(result.alpha) is Allele
    alpha_is_valid = is_class2_pair and type(result.alpha) in {Allele, AlleleWithoutGene, Gene}
    beta_is_allele = is_class2_pair and type(result.beta) is Allele
    beta_is_valid = is_class2_pair and type(result.beta) in {Allele, AlleleWithoutGene, Gene}
    is_allele = t is Allele
    is_allele_without_gene = t is AlleleWithoutGene
    is_serotype = t is Serotype
    is_haplotype = t is Haplotype
    is_gene = t is Gene

    # Penalize serotypes that look like alleles (e.g., B3901, A2403)
    # These should have lower priority than actual allele interpretations
    is_allele_like_serotype = is_serotype and bool(_ALLELE_LIKE_SEROTYPE_PATTERN.match(result.name))

    if is_allele:
        num_allele_fields = result.num_allele_fields
    elif is_class2_pair and alpha_is_allele and beta_is_allele:
        num_allele_fields = max(result.alpha.num_allele_fields, result.beta.num_allele_fields)
    else:
        num_allele_fields = 0

    if hasattr(result, "gene") and result.gene is not None:
        if result.gene.raw_string:
            raw_gene_name_matches_normalized = (
                result.gene.raw_string.lower() == result.gene.name.lower()
            )
        else:
            raw_gene_name_matches_normalized = False
        original_gene_seq_length = len(result.gene.raw_string)
    elif is_gene:
        raw_gene_name_matches_normalized = lower_raw_string == result.name.lower()
        original_gene_seq_length = len(result.raw_string)
    else:
        raw_gene_name_matches_normalized = False
        original_gene_seq_length = 0

    if is_allele:
        allele_fields = result.allele_fields
    elif is_class2_pair and alpha_is_allele and beta_is_allele:
        allele_fields = result.alpha.allele_fields + result.beta.allele_fields
    else:
        allele_fields = ()

    if len(allele_fields) > 0:
        allele_fields_normal = True
        for x in allele_fields:
            if not x.isdigit() or all(c == "0" for c in x) or len(x) < 2 or len(x) > 4:
                allele_fields_normal = False
                break
    else:
        allele_fields_normal = False

    num_alleles_in_haplotype_or_serotype = 0
    if is_serotype or is_haplotype:
        num_alleles_in_haplotype_or_serotype = len(result.alleles)

    return (
        # Penalize serotypes that look like alleles (B3901, A2403, etc.)
        # These should be lower priority than allele interpretations
        not is_allele_like_serotype,
        name_matches_raw_string,
        allele_fields_normal,
        # for Class II pairs, prefer Allele objects for alpha and beta
        # and then any of {AlleleWpithoutGene, Allele, Gene} where
        # genes play the role of mono-morphic alleles
        (is_class2_pair and alpha_is_allele and beta_is_allele),
        (is_class2_pair and alpha_is_valid and beta_is_valid),
        is_allele,
        full_string_matches_raw_string,
        compact_string_matches_raw_string,
        is_gene,
        raw_gene_name_matches_normalized,
        num_allele_fields,
        original_gene_seq_length,
        is_serotype,
        is_haplotype,
        is_allele_without_gene,
        num_alleles_in_haplotype_or_serotype,
        result.raw_string is not None,
        # make sure the ordering is stable by sorting on string
        # representation, even if it's semantically meaningful
        str(result),
    )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sequence
from typing import Union

from .common import unique
from .gene import Gene
from .result_with_species import ResultWithSpecies
from .species import Species


class ResultWithMultipleAlleles(ResultWithSpecies):
    def __init__(
        self,
        species: Species,
        name: str,
        alleles: Sequence[ResultWithSpecies],
        raw_string: Union[str, None] = None,
    ):
        ResultWithSpecies.__init__(self, species=species, raw_string=raw_string)
        self.name = name
        self.alleles = tuple(sorted(alleles))

        genes = []
        for allele in self.alleles:
            # monomorphic alleles sometimes represented just by the gene name
            if type(allele) is Gene:
                genes.append(allele)
            elif allele.has_gene:
                genes.append(allele.gene)
        self.genes = unique(genes)
        inferred_species = unique([gene.species for gene in self.genes])

        if len(inferred_species) == 1:
            inferred_species = inferred_species[0]
            if inferred_species != species:
                raise ValueError(
                    f"Species inferred from given alleles ({inferred_species}) is different from {species}"
                )
        elif len(inferred_species) > 1:
            raise ValueError(
                f"{self.__class__.__name__} '{name}' cannot span genes from multiple species: {inferred_species}"
            )

    @classmethod
    def str_field_names(cls):
        return ("species", "name", "num_alleles")

    @classmethod
    def repr_field_names(cls):
        return ("species", "name", "alleles")

    @classmethod
    def eq_field_names(cls):
        return ("species", "name")

    @property
    def num_alleles(self):
        return len(self.alleles)

    @property
    def gene(self):
        if len(self.genes) == 1:
            return next(iter(self.genes))
        else:
            return None

    @property
    def gene_name(self):
        if self.gene:
            return self.gene.name
        else:
            return None

    def collapse_if_possible(self):
        """
        If this serotype contains a single allele, return it.
        """
        if len(self.alleles) == 1:
            return self.alleles[0]
        return None

    @property
    def has_allele(self):
        return len(self.alleles) > 0

    @property
    def is_class1(self):
        return all(allele.is_class1 for allele in self.alleles)

    @property
    def is_class2(self):
        return all(allele.is_class2 for allele in self.alleles)

    @property
    def is_class2_alpha(self):
        return all(allele.is_class2_alpha for allele in self.alleles)

    @property
    def is_class2_beta(self):
        return all(allele.is_class2_beta for allele in self.alleles)

    @property
    def annotation_null(self):
        return any(allele.annotation_null for allele in self.alleles)

    @property
    def annotation_cystosolic(self):
        return any(allele.annotation_cystosolic for allele in self.alleles)

    @property
    def annotation_secreted(self):
        return any(allele.annotation_secreted for allele in self.alleles)

    @property
    def annotation_questionable(self):
        return any(allele.annotation_questionable for allele in self.alleles)

    @property
    def annotation_low_expression(self):
        return any(allele.annotation_low_expression for allele in self.alleles)

    @property
    def annotation_aberrant_expression(self):
        return any(allele.annotation_aberrant_expression for allele in self.alleles)

    @property
    def annotation_group(self):
        # designates a group of genomic sequence alleles
        # with identical peptide binding region
        return any(allele.annotation_group for allele in self.alleles)

    @property
    def annotation_pseudogene(self):
        # designates a group of genomic sequence alleles
        # with identical peptide binding region
        return any(allele.annotation_pseudogene for allele in self.alleles)

    @property
    def annotation_splice_variant(self):
        # designates a group of genomic sequence alleles
        # with identical peptide binding region
        return any(allele.annotation_splice_variant for allele in self.alleles)

    def restrict_allele_fields(self, num_fields, drop_annotations=False, drop_mutations=False):
        alleles = [
            allele.restrict_allele_fields(num_fields, drop_annotations, drop_mutations)
            for allele in self.alleles
        ]
        return self.copy(alleles=alleles)



from mhcgnomes import Allele, Gene, Pair, Species, parse

from .common import eq_


def test_parse_trout_species_Onmy():
    expected = Species.get("Onmy")
    assert expected is not None
    result = parse("Onmy", raise_on_error=True)
    assert result is not None
    eq_(result, expected)


def test_parse_trout_gene_Onmy_DAB():
    result = parse("Onmy-DAB", raise_on_error=True)
    expected = Gene.get("Onmy", "DAB")
    assert expected is not None
    eq_(result, expected)


def test_parse_trout_allele_Onmy_DAB_0501():
    result = parse("Onmy-DAB*0501", raise_on_error=True)
    expected = Allele.get("Onmy", "DAB", ["05", "01"])
    assert expected is not None
    eq_(result, expected)


def test_parse_trout_allele_Onmy_DAB_050101():
    result = parse("Onmy-DAB*050101", raise_on_error=True)
    expected = Allele.get("Onmy", "DAB", ["05", "01", "01"])
    assert expected is not None
    eq_(result, expected)


def test_parse_grass_carp_B2M_1_ii_star_sep():
    expected = Allele.get("Ctid", "B2M-1", "ii")
    result = parse("Ctid-B2M-1*ii")
    eq_(result, expected)


def test_parse_grass_carp_B2M_1_ii_dash_sep():
    expected = Allele.get("Ctid", "B2M-1", "ii")
    result = parse("Ctid-B2M-1-ii")
    eq_(result, expected)


def test_parse_grass_carp_mhc1_pair_UAA_B2M_1_ii_dash_sep():
    expected = Pair(alpha=Gene.get("Ctid", "UAA"), beta=Allele.get("Ctid", "B2M-1", "ii"))
    result = parse("Ctid-UAA/B2M-1-ii")
    eq_(result, expected)


def test_parse_grass_carp_mhc1_pair_UAA_B2M_1_ii_star_sep():
    expected = Pair(alpha=Gene.get("Ctid", "UAA"), beta=Allele.get("Ctid", "B2M-1", "ii"))
    result = parse("Ctid-UAA/B2M-1*ii")
    eq_(result, expected)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable
from typing import Union

from .mutation import Mutation
from .result_with_mhc_class import ResultWithMhcClass
from .species import Species


class Gene(ResultWithMhcClass):
    """
    Represents an MHC gene within a species.

    A Gene represents a specific MHC locus like HLA-A, HLA-B, or HLA-DRB1.
    It contains species information and can optionally include mutations.

    Parameters
    ----------
    species : Species
        The species this gene belongs to.
    name : str
        The gene name (e.g., "A", "B", "DRB1").
    mutations : Iterable[Mutation], optional
        Any mutations from the reference sequence.
    raw_string : str, optional
        The original unparsed string.

    Attributes
    ----------
    species : Species
        The species this gene belongs to.
    name : str
        The gene name.
    mutations : Iterable[Mutation]
        Any mutations from the reference sequence.
    mhc_class : str
        The MHC class (e.g., "Ia", "Ib", "II", "IIa", "IIb").

    Examples
    --------
    >>> gene = Gene.get("HLA", "A")
    >>> gene.name
    'A'
    >>> gene.mhc_class
    'Ia'
    >>> gene.to_string()
    'HLA-A'
    """

    def __init__(
        self,
        species: Species,
        name: str,
        mutations: Iterable[Mutation] = (),
        raw_string: Union[str, None] = None,
    ):
        ResultWithMhcClass.__init__(
            self,
            species=species,
            mhc_class=species.get_mhc_class_of_gene(name),
            raw_string=raw_string,
        )
        self.name = name
        self.mutations = mutations

    def __hash__(self):
        return hash((self.species, self.name, self.mutations))

    def __eq__(self, other):
        if type(other) is not Gene:
            return False
        return (
            self.species == other.species
            and self.name == other.name
            and self.mutations == other.mutations
        )

    @property
    def raw_string_was_alias(self):
        return self.raw_string in self.species.gene_aliases

    @property
    def gene_name(self):
        return self.name

    @property
    def gene(self):
        return self

    @classmethod
    def get(
        cls, species_prefix: Union[str, Species], gene_name: str, mutations: Iterable[Mutation] = ()
    ):
        """
        Get or create a Gene from species prefix and gene name.

        This is the primary factory method for creating Gene objects.
        It validates that the gene exists in the species ontology.

        Parameters
        ----------
        species_prefix : str or Species
            Species prefix string (e.g., "HLA") or Species object.
        gene_name : str
            The gene name to look up (e.g., "A", "B", "DRB1").
        mutations : Iterable[Mutation], optional
            Any mutations to associate with this gene.

        Returns
        -------
        Gene or None
            The Gene object if found in ontology, otherwise None.

        Examples
        --------
        >>> gene = Gene.get("HLA", "A")
        >>> gene is not None
        True
        >>> gene.name
        'A'
        >>> Gene.get("HLA", "INVALID")  # Returns None
        """
        # use the canonical gene name e.g. "A" and not "a"
        if type(species_prefix) is Species:
            species = species_prefix
        else:
            species = Species.get(species_prefix)
        if species is None:
            return None

        raw_string = gene_name

        gene_name = species.find_matching_gene_name(gene_name)

        if gene_name is None:
            return None

        return Gene(species, gene_name, mutations=mutations, raw_string=raw_string)

    def copy_without_mutations(self):
        return self.copy(mutations=())

    def copy_with_extra_mutations(self, mutations: Iterable[Mutation]):
        return self.copy(mutations=self.mutations + tuple(mutations))

    def mutation_string(self):
        return " ".join([mut.to_string() for mut in self.mutations])

    def to_string(self, include_species=True, use_old_species_prefix=False):
        if include_species:
            # non-classical genes located outside the MHC locus
            # get identified with the common species name if possible
            if self.mhc_class in {"Id"}:
                species_str = self.common_species_name
            elif use_old_species_prefix:
                species_str = self.species.historic_mhc_prefix
            else:
                species_str = self.species_prefix
            result = f"{species_str}-{self.gene_name}"
        else:
            result = self.gene_name
        if self.is_mutant:
            result += f" {self.mutation_string()} mutant"
        return result

    def compact_string(self, include_species=False, use_old_species_prefix=False):
        """
        Compact representation of a Locus, currently same as the
        normalized representation.
        """
        return Gene.to_string(
            self, include_species=include_species, use_old_species_prefix=use_old_species_prefix
        )

    @property
    def is_mutant(self):
        return len(self.mutations) > 0

    def to_record(self):
        """
        Returns a user-viewable ordered dictionary with a representation  of
        this gene, and the values of the following methods:
            - is_mutant
            - get_mhc_class
        """
        d = self.species.to_record()
        d["gene"] = self.to_string()
        d["mhc_class"] = self.mhc_class
        d["mutations"] = self.mutation_string()
        d["is_mutant"] = self.is_mutant
        return d

    def restrict_allele_fields(self, num_fields, drop_annotations=False, drop_mutations=False):
        return self.copy(mutations=[] if drop_mutations else self.mutations)



from mhcgnomes import Allele, Pair, parse

from .common import eq_


def test_normalized_string_hla_a0201_complete():
    name = "HLA-A*02:01"
    expected = "HLA-A*02:01"
    eq_(parse(name).to_string(), expected)


def test_normalized_string_hla_a0201_no_fieldsep():
    name = "HLA-A*0201"
    expected = "HLA-A*02:01"
    eq_(parse(name).to_string(), expected)


def test_normalized_string_hla_a0201_no_species_no_star_no_fieldsep():
    name = "A0201"
    expected = "HLA-A*02:01"
    eq_(parse(name).to_string(), expected)


def test_normalized_string_hla_a0201_no_star_no_fieldsep():
    name = "HLA-A0201"
    expected = "HLA-A*02:01"
    eq_(parse(name).to_string(), expected)


def test_normalized_string_hla_a0201_no_fieldsep_lowercase():
    name = "hla-a*0201"
    expected = "HLA-A*02:01"
    eq_(parse(name).to_string(), expected)


def test_normalized_string_hla_a0201_no_species_no_fieldsep_lowercase():
    name = "a*0201"
    expected = "HLA-A*02:01"
    eq_(parse(name).to_string(), expected)


def test_normalized_string_hla_a0201_no_species_lowercase():
    name = "a*02:01"
    expected = "HLA-A*02:01"
    eq_(parse(name).to_string(), expected)


def test_normalized_string_hla_a0201_no_species_no_star_no_fieldsep_lowercase():
    name = "a0201"
    expected = "HLA-A*02:01"
    eq_(parse(name).to_string(), expected)


def test_normalized_string_hla_a0201_no_species_no_fieldsep():
    name = "A*0201"
    expected = "HLA-A*02:01"
    eq_(parse(name).to_string(), expected)


def test_normalized_string_hla_a0201_no_species():
    name = "A*02:01"
    expected = "HLA-A*02:01"
    eq_(parse(name).to_string(), expected)


hla_02_01_names = [
    "HLA-A*02:01",
    "HLA-A*0201",
    "A*02:01",
    "A*0201",
    "HLA-A02:01",
    # no gene sep
    "HLA-A0201",
    "A0201",
    # lower case
    "hla-a*0201",
    "a*0201",
    "a*02:01",
    "a0201",
]


def test_hla_short_names():
    expected = "A0201"
    for name in hla_02_01_names:
        result = parse(name).compact_string()
        eq_(result, expected)


def test_hla_A2_serotype():
    eq_(parse("A2").compact_string(), "A2")
    eq_(parse("A2").to_string(), "HLA-A2")

    eq_(parse("HLA-A2").compact_string(), "A2")
    eq_(parse("HLA-A2").to_string(), "HLA-A2")


def test_hla_with_3_digit_allele_code():
    # B*15:120
    eq_(parse("HLA-B*15:120").to_string(), "HLA-B*15:120")
    eq_(parse("HLA-B*15:120").compact_string(), "B15120")
    eq_(parse("B15120").to_string(), "HLA-B*15:120")
    eq_(parse("B15120").compact_string(), "B15120")

    # A*02*123
    eq_(parse("HLA-A*02:123").to_string(), "HLA-A*02:123")
    eq_(parse("HLA-A*02:123").compact_string(), "A02123")
    eq_(parse("A02123").to_string(), "HLA-A*02:123")
    eq_(parse("A02123").compact_string(), "A02123")


def test_A_02_01_01_01():
    result = parse("A*02:01:01:01")
    eq_(result.species_prefix, "HLA")
    eq_(result.gene_name, "A")
    eq_(result.allele_fields[0], "02")
    eq_(result.allele_fields[1], "01")
    eq_(result.allele_fields[2], "01")
    eq_(result.allele_fields[3], "01")


def test_parse_human_class2_DRB1_01_02():
    expected = Pair.get(
        Allele.get("HLA", "DRA", "01", "01"),
        Allele.get("HLA", "DRB1", "01", "02"),
    )
    for name in [
        "DRB1_0102",
        "DRB101:02",
        "HLA-DRB1_0102",
        "DRB10102",
        "DRB1*0102",
        "HLA-DRB1*0102",
        "HLA-DRB1*01:02",
        "DRB0102",
    ]:
        parse_result = parse(name, infer_class2_pairing=True)
        eq_(parse_result, expected)


def test_normalized_string_human_class2_DRB1_01_02():
    expected = "HLA-DRA*01:01/DRB1*01:02"
    for name in [
        "DRB1_0102",
        "DRB101:02",
        "HLA-DRB1_0102",
        "DRB10102",
        "DRB1*0102",
        "HLA-DRB1*0102",
        "HLA-DRB1*01:02",
        "DRB0102",
    ]:
        eq_(parse(name, infer_class2_pairing=True).to_string(), expected)


def test_compact_string_string_human_class2_DRB1_01_02():
    expected_compact = "DRB1*0102"
    for name in [
        "DRB1_0102",
        "DRB101:02",
        "HLA-DRB1_0102",
        "DRB10102",
        "DRB1*0102",
        "HLA-DRB1*0102",
        "HLA-DRB1*01:02",
        "DRB0102",
    ]:
        eq_(parse(name).compact_string(), expected_compact)


def test_parse_human_class2_alpha_beta_DPA1_01_05_DPB1_100_01_upper_compact():
    expected = Pair.get(
        Allele.get("HLA", "DPA1", "01", "05"), Allele.get("HLA", "DPB1", "100", "01")
    )
    s = "DPA10105-DPB110001"
    eq_(parse(s, infer_class2_pairing=True), expected)


def test_parse_human_class2_alpha_beta_DPA1_01_05_DPB1_100_01_full_syntax_hyphens():
    expected = Pair.get(
        Allele.get("HLA", "DPA1", "01", "05"), Allele.get("HLA", "DPB1", "100", "01")
    )
    s = "HLA-DPA1*01:05/DPB1*100:01"
    eq_(parse(s, infer_class2_pairing=True), expected)


def test_parse_human_class2_alpha_beta_DPA1_01_05_DPB1_100_01_full_syntax_hyphens_lower():
    expected = Pair.get(
        Allele.get("HLA", "DPA1", "01", "05"), Allele.get("HLA", "DPB1", "100", "01")
    )
    s = "hla-dpa1*0105-dpb1*10001"
    eq_(parse(s, infer_class2_pairing=True), expected)


def test_parse_human_class2_alpha_beta_DPA1_01_05_DPB1_100_01_no_hla_lower_no_colons():
    expected = Pair.get(
        Allele.get("HLA", "DPA1", "01", "05"), Allele.get("HLA", "DPB1", "100", "01")
    )
    s = "dpa1*0105-dpb1*10001"
    eq_(parse(s, infer_class2_pairing=True), expected)


def test_parse_human_class2_alpha_beta_DPA1_01_05_DPB1_100_01_full_syntax_slash():
    expected = Pair.get(
        Allele.get("HLA", "DPA1", "01", "05"), Allele.get("HLA", "DPB1", "100", "01")
    )
    s = "HLA-DPA1*01:05/DPB1*100:01"
    eq_(parse(s, infer_class2_pairing=True), expected)


def test_parse_human_class2_alpha_beta_DPA1_01_05_DPB1_100_01_compact_slash():
    expected = Pair.get(
        Allele.get("HLA", "DPA1", "01", "05"), Allele.get("HLA", "DPB1", "100", "01")
    )
    s = "DPA10105/DPB110001"
    eq_(parse(s, infer_class2_pairing=True), expected)


def test_parse_all_parameters_true_human_class2_alpha_beta_DPA1_01_05_DPB1_100_01():
    expected = Pair.get(
        Allele.get("HLA", "DPA1", "01", "05"), Allele.get("HLA", "DPB1", "100", "01")
    )
    for name in [
        "DPA10105-DPB110001",
        "HLA-DPA1*01:05-DPB1*100:01",
        "hla-dpa1*0105-dpb1*10001",
        "dpa1*0105-dpb1*10001",
        "HLA-DPA1*01:05/DPB1*100:01",
        "DPA10105/DPB110001",
    ]:
        parse_result = parse(name, use_allele_aliases=True, infer_class2_pairing=True)
        eq_(parse_result, expected)


def test_normalize_string_human_class2_alpha_beta_DPA1_01_05_DPB1_100_01():
    expected = "HLA-DPA1*01:05/DPB1*100:01"
    for name in [
        "DPA10105-DPB110001",
        "HLA-DPA1*01:05-DPB1*100:01",
        "hla-dpa1*0105-dpb1*10001",
        "dpa1*0105-dpb1*10001",
        "HLA-DPA1*01:05/DPB1*100:01",
        "DPA10105/DPB110001",
    ]:
        eq_(parse(name, infer_class2_pairing=True).to_string(), expected)


def test_compact_string_human_class2_alpha_beta_DPA1_01_05_DPB1_100_01():
    expected_compact = "DPA1*0105-DPB1*10001"
    for name in [
        "DPA10105-DPB110001",
        "HLA-DPA1*01:05-DPB1*100:01",
        "hla-dpa1*0105-dpb1*10001",
        "dpa1*0105-dpb1*10001",
        "HLA-DPA1*01:05/DPB1*100:01",
        "DPA10105/DPB110001",
    ]:
        eq_(parse(name, infer_class2_pairing=True).compact_string(), expected_compact)


def test_alpha_chain_inference_DP():
    # example of common DP haplotype from wikipedia article on HLA-DP
    just_beta = "HLA-DPB1*04:01"
    expected = "HLA-DPA1*01:03/DPB1*04:01"
    eq_(parse(just_beta, infer_class2_pairing=True).to_string(), expected)
    eq_(parse(just_beta, infer_class2_pairing=False).to_string(), just_beta)


def test_alpha_chain_inference_DQ():
    # example of common DQ haplotype from wikipedia article on HLA-DQ
    just_beta = "HLA-DQB1*06:02"
    expected = "HLA-DQA1*01:02/DQB1*06:02"
    eq_(parse(just_beta, infer_class2_pairing=True).to_string(), expected)
    eq_(parse(just_beta, infer_class2_pairing=False).to_string(), just_beta)


def test_human_class2_pair_with_mutation():
    allele = "HLA-DRA*01:01/DRB1*01:01 G86Y mutant"
    result = parse(allele)
    eq_(type(result), Pair)
    eq_(result.beta.is_mutant, True)


def test_human_class2_beta_with_mutation():
    allele = "DRB1*01:01 G86Y mutant"
    result = parse(allele, infer_class2_pairing=True)
    eq_(type(result), Pair)
    eq_(result.beta.is_mutant, True)


def test_DRB5_0108N():
    allele = "DRB5_0108N"
    result = parse(allele)
    assert result is not None
    eq_(type(result), Allele)
    assert result.annotation_null


def test_DRB5_0108N_with_inferred_alpha():
    allele = "DRB5_0108N"
    result = parse(allele, infer_class2_pairing=True)
    assert result is not None
    eq_(type(result), Pair)
    assert result.annotation_null



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sequence
from typing import Union

from .allele import Allele
from .class2_locus import Class2Locus
from .mhc_class_helpers import (
    is_valid_restriction,
    restrict_alleles,
)
from .pair import Pair
from .result_with_multiple_alleles import ResultWithMultipleAlleles
from .species import Species


class Haplotype(ResultWithMultipleAlleles):
    """
    Represents a named MHC haplotype containing multiple alleles.

    A Haplotype is a collection of MHC alleles that are inherited together
    on a single chromosome. Common examples include mouse haplotypes like
    "H2-k" which define the complete set of MHC alleles for that strain.

    Haplotypes can be restricted by MHC class (I or II) or by a specific
    Class II locus (e.g., DR, DQ, DP).

    Parameters
    ----------
    species : Species
        The species this haplotype belongs to.
    name : str
        The haplotype name (e.g., "k", "d", "b" for mouse).
    alleles : Sequence[Allele]
        The alleles contained in this haplotype.
    class_restriction : str, optional
        MHC class restriction ("I" or "II").
    locus_restriction : Class2Locus, optional
        Specific Class II locus restriction.
    parent_haplotypes : Sequence[Haplotype], optional
        Parent haplotypes if this is derived from others.
    raw_string : str, optional
        The original unparsed string.

    Attributes
    ----------
    species : Species
        The species this haplotype belongs to.
    name : str
        The haplotype name.
    alleles : Sequence[Allele]
        The alleles in this haplotype.
    class_restriction : str or None
        MHC class restriction if any.
    locus_restriction : Class2Locus or None
        Class II locus restriction if any.

    Examples
    --------
    >>> from mhcgnomes import parse
    >>> hap = parse("H2-b")
    >>> hap.name
    'b'
    >>> hap.species.common_name
    'mouse'
    """

    def __init__(
        self,
        species: Species,
        name: str,
        alleles: Sequence[Allele],
        class_restriction: Union[str, None] = None,
        locus_restriction: Union[Class2Locus, None] = None,
        parent_haplotypes: Union[Sequence["Haplotype"], None] = None,
        raw_string: Union[str, None] = None,
    ):
        ResultWithMultipleAlleles.__init__(
            self, species=species, name=name, alleles=alleles, raw_string=raw_string
        )
        self.class_restriction = class_restriction
        self.locus_restriction = locus_restriction
        self.parent_haplotypes = parent_haplotypes

    @classmethod
    def str_field_names(cls):
        return ("species", "name", "num_alleles", "class_restriction", "locus_restriction")

    @classmethod
    def repr_field_names(cls):
        return ("species", "name", "alleles", "class_restriction", "locus_restriction")

    @classmethod
    def eq_field_names(cls):
        return ("species", "name", "class_restriction", "locus_restriction")

    @property
    def haplotype_name(self):
        return self.name

    def restrict_mhc_class(self, class_restriction, raise_on_error=True):
        if class_restriction is None:
            return self
        if self.class_restriction == class_restriction:
            return self
        if not is_valid_restriction(self.class_restriction, class_restriction):
            if raise_on_error:
                raise ValueError(
                    f"Cannot restrict '{self.to_string()}' to class '{class_restriction}'"
                )
            else:
                return None
        restricted_alleles = restrict_alleles(self.alleles, class_restriction)
        return Haplotype(
            species=self.species,
            name=self.name,
            alleles=restricted_alleles,
            class_restriction=class_restriction,
            locus_restriction=self.locus_restriction,
            raw_string=self.raw_string,
        )

    def restrict_class2_locus(self, class2_locus: Class2Locus, raise_on_error=True):
        if class2_locus is None:
            return self
        if self.locus_restriction == class2_locus:
            return self
        if self.locus_restriction is not None:
            if raise_on_error:
                raise ValueError(
                    f"Haplotype {self} already has locus restriction, cannot restrict to {class2_locus}"
                )
            else:
                return None
        valid_genes = set(class2_locus.genes)

        restricted_alleles = [allele for allele in self.alleles if allele.gene in valid_genes]
        return Haplotype(
            species=self.species,
            name=self.name,
            alleles=restricted_alleles,
            class_restriction=self.class_restriction,
            locus_restriction=class2_locus,
            raw_string=self.raw_string,
        )

    def collapse_if_possible(self):
        """
        If this haplotype contains a single allele or a single pair of
        class alleles for the same locus, return it. Otherwise return None.
        """
        allele = self.get_single_allele()
        if allele:
            return allele

        class2_pair = self.get_class2_pair()
        if class2_pair:
            return class2_pair
        return None

    def get_single_allele(self):
        if len(self.alleles) == 1:
            return self.alleles[0]
        return None

    def get_class2_pair(self):
        if self.locus_restriction is None:
            return None
        alpha_genes = self.locus_restriction.alpha_chain_genes
        beta_genes = self.locus_restriction.beta_chain_genes

        if len(alpha_genes) == 0 or len(beta_genes) == 0:
            return None
        alpha_alleles = [a for a in self.alleles if a.gene in alpha_genes]
        beta_alleles = [a for a in self.alleles if a.gene in beta_genes]
        if len(alpha_alleles) != 1 or len(beta_alleles) != 1:
            return None

        alpha = alpha_alleles[0]
        beta = beta_alleles[0]
        return Pair(alpha, beta)

    def to_string(self, include_species=True, use_old_species_prefix=False):
        if self.locus_restriction:
            prefix = self.locus_restriction.to_string(
                include_species=include_species, use_old_species_prefix=use_old_species_prefix
            )
        else:
            prefix = self.species.to_string(
                include_species=include_species, use_old_species_prefix=use_old_species_prefix
            )

        if prefix:
            result = f"{prefix}-{self.name}"
        else:
            result = self.name

        if self.class_restriction:
            result += f" class {self.class_restriction}"

        return result

    def compact_string(self, include_species=False):
        return self.to_string(include_species=include_species)



from collections import defaultdict

from mhcgnomes.normalizing_set import NormalizingSet

from .common import eq_


def test_NormalizingSet_get_original():
    s = NormalizingSet()
    item = "AB:cd"
    s.add(item)
    original = s.get_original("abcd")
    eq_(original, item)


def test_NormalizingSet_eq():
    a = NormalizingSet()
    a.add("yo")
    b = NormalizingSet()
    b.add("YO ")
    eq_(a, b)


def test_NormalizingSet_in_defaultdict():
    d = defaultdict(NormalizingSet)
    d[1].add("YO ")
    expected = NormalizingSet()
    expected.add("yo")
    eq_(d[1], expected)



from mhcgnomes import Allele, parse

from .common import eq_

"""
Collection of tests for UniProt descriptions of MHC alleles
"""


def test_parse_uniprot_MOUSE_MHC_class_I_L_q_alpha_chain():
    s = "MOUSE MHC class I L-q alpha-chain"
    result = parse(s)
    eq_(result, Allele.get("H2", "L", "q"))


def test_parse_uniprot_H2_class_I_histocompatibility_antigen_K_B_alpha_chain():
    s = "H-2 class I histocompatibility antigen, K-B alpha chain"
    result = parse(s)
    eq_(result, Allele.get("H2", "K", "b"))


def test_parse_uniprot_H2_class_I_histocompatibility_antigen_K_Q_alpha_chain_fragment():
    s = "H-2 class I histocompatibility antigen, K-Q alpha chain (Fragment)"
    result = parse(s)
    eq_(result, Allele.get("H2", "K", "q"))


def test_parse_uniprot_MHC_class_I_histocompatibility_antigen_L_Q_alpha_chain_fragment_with_attributes():
    s = "MHC class I L-q alpha-chain gene (H2 haplotype), exons 1-3 (Fragment) OS=Mus musculus OX=10090 PE=3 SV=1"
    result = parse(s)
    eq_(result, Allele.get("H2", "L", "q"))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .result import Result


class ResultWithSpecies(Result):
    """
    Common base class for any result object which has a species field.

    Useful for sharing helper methods that rely on the 'species' field.
    """

    def __init__(self, species, raw_string=None):
        Result.__init__(self, raw_string=raw_string)
        self.species = species

    @property
    def has_species(self):
        return True

    @property
    def species_prefix(self):
        return self.species.prefix

    @property
    def species_common_name(self):
        return self.species.common_species_name

    @property
    def common_species_name(self):
        return self.species.common_species_name

    @property
    def is_mouse(self):
        return self.species.is_mouse

    @property
    def is_chicken(self):
        return self.species.is_chicken

    @property
    def is_rat(self):
        return self.species.is_rat

    @property
    def is_human(self):
        return self.species.is_human

    @property
    def is_dog(self):
        return self.species.is_dog

    @property
    def is_cat(self):
        return self.species.is_cat

    @property
    def is_pig(self):
        return self.species.is_pig

    @property
    def is_cow(self):
        return self.species.is_cow

    @property
    def is_horse(self):
        return self.species.is_horse



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from serializable import Serializable

from .common import cache
from .parsing_helpers import strip_chars
from .token import Token
from .token_substitution import simplify_tokens


class TokenizationResult(Serializable):
    def __init__(self, tokens, ignored_tokens, attributes, raw_string, trimmed_string):
        self.tokens = tuple(tokens)
        self.ignored_tokens = tuple(ignored_tokens)
        self.attributes = attributes
        self.raw_string = raw_string
        self.trimmed_string = trimmed_string


def deparen(s):
    """
    Remove all interior parantheses from sequence.
    """
    return s.replace("(", "").replace(")", "")


def normalize_token(s):
    """
    Apply all string transformations to turn non-whitespace character sequence
    into a token.
    """
    return strip_chars(deparen(s.lower()), "-, ").strip()


def split_token_sequences(s, token_seps="/"):
    """
    Split string on whitespace. Also split_token_sequences on slashes but preserve
    them in the token sequence.
    """
    results = s.split()
    for sep in token_seps:
        new_results = []
        for part in results:
            for i, sub_part in enumerate(part.split(sep)):
                if i > 0:
                    new_results.append(sep)
                new_results.append(sub_part)
        results = new_results

    return results


def split_and_extract_attributes(s):
    parts = split_token_sequences(s)
    if "=" not in s:
        return parts, {}

    parts_before_attributes = []
    attributes = {}
    found_first_attribute = False
    attribute_key_in_progress = None
    attribute_values_in_progress = None

    for part in parts:
        if part.count("=") == 1:
            found_first_attribute = True
            if attribute_key_in_progress:
                attributes[attribute_key_in_progress] = " ".join(attribute_values_in_progress)
            attribute_key_in_progress, first_value = part.split("=")
            attribute_values_in_progress = [first_value]
        elif not found_first_attribute:
            parts_before_attributes.append(part)
        else:
            attribute_values_in_progress.append(part)
    if attribute_key_in_progress:
        attributes[attribute_key_in_progress] = " ".join(attribute_values_in_progress)
    return parts_before_attributes, attributes


def strip_whitespace_and_remove_quotes(name: str):
    return name.replace('"', "").replace("'", "").strip()


@cache
def tokenize(name):
    """
    Splits name into token sequence and returns TokenizationResult which
    contains a sequence of Token objects.
    """
    name = str(name)
    trimmed_name = strip_whitespace_and_remove_quotes(name)
    raw_string_parts, attributes = split_and_extract_attributes(trimmed_name)
    token_sequences = [normalize_token(p) for p in raw_string_parts]
    tokens = [
        Token(seq, raw_string) for (seq, raw_string) in zip(token_sequences, raw_string_parts)
    ]
    tokens, ignored_tokens = simplify_tokens(tokens)
    return TokenizationResult(
        tokens=tokens,
        ignored_tokens=ignored_tokens,
        attributes=attributes,
        raw_string=name,
        trimmed_string=trimmed_name,
    )



"""
Tests for HLA Class II heterodimer dot notation parsing.

HLA-DQ molecules are heterodimers consisting of an alpha chain (DQA1) and
a beta chain (DQB1). The shorthand "DQx.y" notation encodes both:
- x = the DQ serotype number (determined by beta chain)
- y = the alpha chain variant number

For example, DQ2.5 means "DQ2 serotype with alpha-5 chain" = DQA1*05:01/DQB1*02:01.

Note: HLA-DR does NOT use this notation because DRA is essentially monomorphic
(non-variable), so DR serotypes are determined solely by DRB alleles.

References:
-----------
DQ2.5 (DQA1*05:01/DQB1*02:01):
    - Lundin KE, et al. "Gliadin-specific, HLA-DQ restricted T cells from coeliac
      disease patients." PMC3872820 (2014)
      https://pmc.ncbi.nlm.nih.gov/articles/PMC3872820/
    - Megiorni F, Pizzuti A. "HLA-DQA1 and HLA-DQB1 in Celiac disease
      predisposition." J Biomed Sci. PMC3482388 (2012)
      https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3482388/

DQ2.2 (DQA1*02:01/DQB1*02:02):
    - Dahal-Koirala S, et al. "A molecular basis for the T cell response in
      HLA-DQ2.2 mediated celiac disease." PNAS 117(6):3063-3070 (2020)
      https://www.pnas.org/doi/10.1073/pnas.1914308117

DQ2.3 (DQA1*03:01/DQB1*02:01):
    - Broughton SE, et al. "Structural and Functional Studies of trans-Encoded
      HLA-DQ2.3 (DQA1*03:01/DQB1*02:01) Protein Molecule." PMC3340161 (2012)
      https://pmc.ncbi.nlm.nih.gov/articles/PMC3340161/

DQ7.5 (DQA1*05:05/DQB1*03:01):
    - Mayo Clinic Laboratories. "Celiac Associated HLA-DQ Alpha 1 and DQ Beta 1
      DNA Typing." Test ID: CELI.
      https://gi.testcatalog.org/show/CELI

DQ3.1 (DQA1*03:01/DQB1*03:01):
    - Sidney J, et al. "Definition of a DQ3.1-specific binding motif."
      J Immunol. 1994 Oct 1;153(7):3194-203. PMID: 7512598
      https://pubmed.ncbi.nlm.nih.gov/7512598/
    - IEDB assay data referencing HLA-DQA1*03:01/DQB1*03:01

DQ8.5 (DQA1*05:01/DQB1*03:02):
    - Petersen J, et al. "Diverse T Cell Receptor Gene Usage in HLA-DQ8-Associated
      Celiac Disease Converges into a Consensus Binding Solution."
      Structure. 2016 Oct 4;24(10):1643-1657. PMID: 27568928
      https://pubmed.ncbi.nlm.nih.gov/27568928/
    - Crystal structures: PDB 5KSA, 5KSB

IEDB occurrence counts (mhc_ligand_full.csv):
    DQ2.2:  45,950 occurrences
    DQ7.5:  45,944 occurrences
    DQ2.5:  43,167 occurrences
    DR11.1:     84 occurrences (T cell clone name, NOT a heterodimer)
    DQ3.1:      47 occurrences
    DQ2.3:      14 occurrences
    DQ8.5:       2 occurrences
"""

import pytest

from mhcgnomes import parse
from mhcgnomes.errors import ParseError
from mhcgnomes.pair import Pair


class TestHeterodimersBasicParsing:
    """Test that heterodimer shorthand notation parses to Pair objects."""

    @pytest.mark.parametrize(
        "shorthand,expected_alpha,expected_beta",
        [
            # DQ2.5 - most common celiac disease haplotype (43,167 in IEDB)
            # Source: PMC3872820, PMC3482388
            ("DQ2.5", "DQA1*05:01", "DQB1*02:01"),
            ("HLA-DQ2.5", "DQA1*05:01", "DQB1*02:01"),
            # DQ2.2 - second DQ2 variant (45,950 in IEDB)
            # Source: PNAS 1914308117
            ("DQ2.2", "DQA1*02:01", "DQB1*02:02"),
            ("HLA-DQ2.2", "DQA1*02:01", "DQB1*02:02"),
            # DQ2.3 - trans-encoded variant associated with T1D
            # Source: PMC3340161
            ("DQ2.3", "DQA1*03:01", "DQB1*02:01"),
            ("HLA-DQ2.3", "DQA1*03:01", "DQB1*02:01"),
            # DQ7.5 - shares alpha chain with DQ2.5 (45,944 in IEDB)
            # Source: Mayo Clinic Labs CELI test
            ("DQ7.5", "DQA1*05:05", "DQB1*03:01"),
            ("HLA-DQ7.5", "DQA1*05:05", "DQB1*03:01"),
            # DQ3.1 - well-characterized binding motif (47 in IEDB)
            # Source: J Immunol 1994 PMID:7512598
            ("DQ3.1", "DQA1*03:01", "DQB1*03:01"),
            ("HLA-DQ3.1", "DQA1*03:01", "DQB1*03:01"),
            # DQ8.5 - trans-encoded DQ8 variant (2 in IEDB)
            # Source: Structure 2016 PMID:27568928, PDB 5KSA/5KSB
            ("DQ8.5", "DQA1*05:01", "DQB1*03:02"),
            ("HLA-DQ8.5", "DQA1*05:01", "DQB1*03:02"),
        ],
    )
    def test_heterodimer_parses_to_pair(self, shorthand, expected_alpha, expected_beta):
        """Heterodimer shorthand should parse to a Pair with correct alpha/beta alleles."""
        result = parse(shorthand)
        assert isinstance(result, Pair), f"{shorthand} should parse as Pair, got {type(result)}"

        # Verify alpha chain
        alpha_str = result.alpha.to_string()
        assert expected_alpha in alpha_str, (
            f"{shorthand} alpha chain should be {expected_alpha}, got {alpha_str}"
        )

        # Verify beta chain
        beta_str = result.beta.to_string()
        assert expected_beta in beta_str, (
            f"{shorthand} beta chain should be {expected_beta}, got {beta_str}"
        )


class TestHeterodimersStringOutput:
    """Test that heterodimer Pair objects produce correct string representations."""

    @pytest.mark.parametrize(
        "shorthand,expected_compact",
        [
            ("DQ2.5", "HLA-DQA1*05:01/DQB1*02:01"),
            ("DQ2.2", "HLA-DQA1*02:01/DQB1*02:02"),
            ("DQ2.3", "HLA-DQA1*03:01/DQB1*02:01"),
            ("DQ7.5", "HLA-DQA1*05:05/DQB1*03:01"),
            ("DQ3.1", "HLA-DQA1*03:01/DQB1*03:01"),
            ("DQ8.5", "HLA-DQA1*05:01/DQB1*03:02"),
        ],
    )
    def test_heterodimer_to_string(self, shorthand, expected_compact):
        """Heterodimer should produce standard pair notation when converted to string."""
        result = parse(shorthand)
        assert result.to_string() == expected_compact


class TestHeterodimersPreserveRawString:
    """Test that parsing preserves the original input string."""

    @pytest.mark.parametrize(
        "shorthand",
        [
            "DQ2.5",
            "HLA-DQ2.5",
            "DQ2.2",
            "DQ7.5",
        ],
    )
    def test_raw_string_preserved(self, shorthand):
        """The original input string should be preserved in raw_string."""
        result = parse(shorthand)
        assert result.raw_string == shorthand


class TestHeterodimersVsSerotypes:
    """Test that heterodimer notation is distinct from serotype notation."""

    def test_dq2_is_serotype(self):
        """DQ2 (without dot) should parse as Serotype, not Pair."""
        from mhcgnomes.serotype import Serotype

        result = parse("DQ2")
        assert isinstance(result, Serotype)
        assert result.name == "DQ2"

    def test_dq8_is_serotype(self):
        """DQ8 (without dot) should parse as Serotype."""
        from mhcgnomes.serotype import Serotype

        result = parse("DQ8")
        assert isinstance(result, Serotype)
        assert result.name == "DQ8"

    def test_dq2_5_is_pair(self):
        """DQ2.5 (with dot) should parse as Pair, not Serotype."""
        result = parse("DQ2.5")
        assert isinstance(result, Pair)
        # Should NOT be a Serotype
        from mhcgnomes.serotype import Serotype

        assert not isinstance(result, Serotype)


class TestHeterodimersEquivalence:
    """Test that heterodimer shorthand produces same result as explicit pair notation."""

    @pytest.mark.parametrize(
        "shorthand,explicit",
        [
            ("DQ2.5", "DQA1*05:01/DQB1*02:01"),
            ("DQ2.2", "DQA1*02:01/DQB1*02:02"),
            ("DQ2.3", "DQA1*03:01/DQB1*02:01"),
            ("DQ7.5", "DQA1*05:05/DQB1*03:01"),
            ("DQ3.1", "DQA1*03:01/DQB1*03:01"),
            ("DQ8.5", "DQA1*05:01/DQB1*03:02"),
        ],
    )
    def test_shorthand_equals_explicit_notation(self, shorthand, explicit):
        """Parsing shorthand should produce equivalent result to parsing explicit notation."""
        shorthand_result = parse(shorthand)
        explicit_result = parse(explicit)

        # Both should be Pair objects
        assert isinstance(shorthand_result, Pair)
        assert isinstance(explicit_result, Pair)

        # Alpha chains should match
        assert shorthand_result.alpha == explicit_result.alpha

        # Beta chains should match
        assert shorthand_result.beta == explicit_result.beta


class TestHeterodimersNamingLogic:
    """
    Test the naming logic: DQx.y means DQx serotype with alpha-y chain.

    The notation encodes:
    - First number (x) = DQ serotype (determined by DQB1 allele)
    - Second number (y) = alpha chain variant (DQA1 allele group)

    This is based on the classical serological naming where:
    - DQ2 serotype = DQB1*02:xx alleles
    - DQ3 serotype = DQB1*03:xx alleles
    - DQ7 serotype = DQB1*03:01 specifically (split of DQ3)
    - DQ8 serotype = DQB1*03:02 specifically (split of DQ3)
    """

    def test_dq2_variants_share_beta_chain_family(self):
        """All DQ2.x variants should have DQB1*02:xx beta chains."""
        for shorthand in ["DQ2.2", "DQ2.3", "DQ2.5"]:
            result = parse(shorthand)
            assert isinstance(result, Pair)
            beta_gene = result.beta.gene.name
            assert beta_gene == "DQB1", f"{shorthand} beta should be DQB1, got {beta_gene}"
            beta_group = result.beta.allele_fields[0]
            assert beta_group == "02", f"{shorthand} beta should be *02:xx, got *{beta_group}:xx"

    def test_alpha_5_variants(self):
        """DQx.5 variants should have DQA1*05:xx alpha chains."""
        for shorthand in ["DQ2.5", "DQ7.5", "DQ8.5"]:
            result = parse(shorthand)
            assert isinstance(result, Pair)
            alpha_gene = result.alpha.gene.name
            assert alpha_gene == "DQA1", f"{shorthand} alpha should be DQA1"
            alpha_group = result.alpha.allele_fields[0]
            assert alpha_group == "05", f"{shorthand} alpha should be *05:xx, got *{alpha_group}:xx"


class TestHeterodimersEdgeCases:
    """Test edge cases and unusual inputs."""

    def test_case_insensitive(self):
        """Heterodimer parsing should be case-insensitive."""
        variants = ["DQ2.5", "dq2.5", "Dq2.5", "dQ2.5"]
        results = [parse(v) for v in variants]

        # All should produce equivalent Pair objects
        for i, result in enumerate(results):
            assert isinstance(result, Pair), f"{variants[i]} should parse as Pair"

        # All should have same alpha/beta
        for i in range(1, len(results)):
            assert results[i].alpha == results[0].alpha
            assert results[i].beta == results[0].beta

    def test_invalid_heterodimer_patterns_fail(self):
        """Invalid heterodimer patterns should raise ParseError or not parse as Pair."""
        import contextlib

        invalid_patterns = [
            "DQ2.",  # Missing second number
            "DQ.5",  # Missing first number
            "DQ25",  # No dot (this is ambiguous - could be serotype)
            "DR2.5",  # DR doesn't use this notation (DRA is monomorphic)
        ]
        for pattern in invalid_patterns:
            # These should either fail or parse as something other than heterodimer Pair
            with contextlib.suppress(ParseError):
                result = parse(pattern)
                # If it parses, it shouldn't be a Pair from heterodimer notation
                # (e.g., DQ25 might parse as something else)
                assert not isinstance(result, Pair), f"{pattern} should not parse as Pair"


class TestDRDoesNotUseHeterodimerNotation:
    """
    Test that DR molecules don't use the DQx.y heterodimer notation.

    HLA-DR is different from HLA-DQ:
    - DRA (alpha chain) is essentially monomorphic (no functional variation)
    - All DR serotype specificity comes from DRB (beta chain)
    - Therefore "DRx.y" notation doesn't make sense for DR

    The "DR11.1" found in IEDB (84 occurrences) refers to a T cell clone name,
    not a heterodimer designation.

    Reference:
    - HLA-DRA Wikipedia: "Unlike the alpha chains of other Human MHC class II
      molecules, the alpha subunit is practically invariable."
      https://en.wikipedia.org/wiki/HLA-DRA
    """

    def test_dr11_is_serotype(self):
        """DR11 should parse as serotype (beta chain determines DR specificity)."""
        from mhcgnomes.serotype import Serotype

        result = parse("DR11")
        assert isinstance(result, Serotype)
        assert result.name == "DR11"

    def test_dr11_1_is_not_valid_heterodimer(self):
        """DR11.1 should not parse as a heterodimer (DRA is monomorphic)."""
        # DR11.1 in IEDB refers to a T cell clone, not a molecule
        # This should either fail or not parse as a Pair
        try:
            result = parse("DR11.1")
            assert not isinstance(result, Pair), "DR11.1 should not parse as Pair"
        except ParseError:
            pass  # This is acceptable - DR11.1 is not valid heterodimer notation


class TestCeliacDiseaseAssociatedHeterodimers:
    """
    Test heterodimers commonly referenced in celiac disease literature.

    ~90% of celiac patients carry DQ2.5 or DQ2.2, often in combination.
    ~5-10% carry DQ8 (not dot notation since it's determined by DQB1*03:02 alone).

    The trans-encoded DQ2.5 can form from DQ2.2/DQ7.5 heterozygotes:
    - DQ2.2 provides DQB1*02:02 (similar to DQB1*02:01)
    - DQ7.5 provides DQA1*05:05 (similar to DQA1*05:01)
    - Together they can form DQ2.5-like molecule in trans

    References:
    - Megiorni F, Pizzuti A. J Biomed Sci. 2012. PMC3482388
    - Lundin KE, et al. PMC3872820
    """

    def test_celiac_primary_risk_heterodimer(self):
        """DQ2.5 is the primary celiac disease risk heterodimer."""
        result = parse("DQ2.5")
        assert isinstance(result, Pair)
        # Verify it's the correct molecule
        assert "DQA1" in result.alpha.gene.name
        assert "DQB1" in result.beta.gene.name

    def test_celiac_secondary_risk_heterodimer(self):
        """DQ2.2 confers lower celiac risk than DQ2.5."""
        result = parse("DQ2.2")
        assert isinstance(result, Pair)

    def test_trans_contributing_heterodimer(self):
        """DQ7.5 can contribute to trans-encoded DQ2.5 formation."""
        result = parse("DQ7.5")
        assert isinstance(result, Pair)
        # DQ7.5 shares the alpha chain with DQ2.5
        assert result.alpha.allele_fields[0] == "05"


class TestIEDBOccurrencePatterns:
    """
    Test all heterodimer patterns found in IEDB mhc_ligand_full.csv.

    Occurrence counts:
        DQ2.2:  45,950
        DQ7.5:  45,944
        DQ2.5:  43,167
        DQ3.1:      47
        DQ2.3:      14
        DQ8.5:       2
        DQ2.25:      1 (likely typo)
        DQ2.22:      1 (likely typo)
    """

    @pytest.mark.parametrize(
        "shorthand,count",
        [
            ("DQ2.2", 45950),
            ("DQ7.5", 45944),
            ("DQ2.5", 43167),
            ("DQ3.1", 47),
            ("DQ2.3", 14),
            ("DQ8.5", 2),
        ],
    )
    def test_iedb_heterodimer_patterns(self, shorthand, count):
        """All heterodimer patterns from IEDB should parse as Pair."""
        result = parse(shorthand)
        assert isinstance(result, Pair), (
            f"{shorthand} ({count} occurrences in IEDB) should parse as Pair"
        )

    def test_rare_iedb_variants_are_typos(self):
        """DQ2.22 and DQ2.25 (1 occurrence each) are likely typos."""
        # These appear only once in IEDB and don't follow standard naming
        # DQ2.22 - might be typo for DQ2.2
        # DQ2.25 - might be typo for DQ2.5 or DQ2.2/DQ2.5
        # We don't need to support these
        pass



# Comprehensive tests using real-world HLA allele formats encountered in the wild
#
# These tests cover various API methods on alleles commonly seen in:
# - IEDB database
# - NetMHCpan input/output
# - Clinical HLA typing reports
# - Research publications
# - UniProt descriptions

import pytest

from mhcgnomes import (
    Allele,
    Gene,
    MhcClass,
    Pair,
    ParseError,
    Serotype,
    Species,
    parse,
)

from .common import eq_

# ============================================================================
# Standard HLA Class I alleles - the most common formats
# ============================================================================


class TestStandardClassIAlleles:
    """Tests for standard HLA Class I allele formats."""

    # Common alleles seen in immunology research
    STANDARD_CLASS1_ALLELES = [
        # (input, expected_gene, expected_fields, expected_annotations)
        ("HLA-A*02:01", "A", ("02", "01"), ()),
        ("HLA-A*01:01", "A", ("01", "01"), ()),
        ("HLA-A*03:01", "A", ("03", "01"), ()),
        ("HLA-A*11:01", "A", ("11", "01"), ()),
        ("HLA-A*24:02", "A", ("24", "02"), ()),
        ("HLA-B*07:02", "B", ("07", "02"), ()),
        ("HLA-B*08:01", "B", ("08", "01"), ()),
        ("HLA-B*15:01", "B", ("15", "01"), ()),
        ("HLA-B*27:05", "B", ("27", "05"), ()),
        ("HLA-B*35:01", "B", ("35", "01"), ()),
        ("HLA-B*44:02", "B", ("44", "02"), ()),
        ("HLA-B*44:03", "B", ("44", "03"), ()),
        ("HLA-B*51:01", "B", ("51", "01"), ()),
        ("HLA-B*57:01", "B", ("57", "01"), ()),  # Associated with HIV control
        ("HLA-C*01:02", "C", ("01", "02"), ()),
        ("HLA-C*03:04", "C", ("03", "04"), ()),
        ("HLA-C*04:01", "C", ("04", "01"), ()),
        ("HLA-C*05:01", "C", ("05", "01"), ()),
        ("HLA-C*06:02", "C", ("06", "02"), ()),  # Psoriasis association
        ("HLA-C*07:01", "C", ("07", "01"), ()),
        ("HLA-C*07:02", "C", ("07", "02"), ()),
    ]

    @pytest.mark.parametrize("allele_str,gene,fields,annotations", STANDARD_CLASS1_ALLELES)
    def test_parse_standard_class1(self, allele_str, gene, fields, annotations):
        """Parse standard Class I alleles and verify all properties."""
        result = parse(allele_str)
        assert result is not None, f"Failed to parse {allele_str}"
        assert type(result) is Allele
        eq_(result.gene_name, gene)
        eq_(result.allele_fields, fields)
        eq_(result.annotations, annotations)
        assert result.is_class1
        assert not result.is_class2
        assert result.species.is_human

    @pytest.mark.parametrize("allele_str,gene,fields,annotations", STANDARD_CLASS1_ALLELES)
    def test_to_string_roundtrip(self, allele_str, gene, fields, annotations):
        """Parse and convert back to string - should be normalizable."""
        result = parse(allele_str)
        output = result.to_string()
        # Re-parse the output
        reparsed = parse(output)
        eq_(result, reparsed)

    @pytest.mark.parametrize("allele_str,gene,fields,annotations", STANDARD_CLASS1_ALLELES)
    def test_compact_string(self, allele_str, gene, fields, annotations):
        """Test compact string output without separators."""
        result = parse(allele_str)
        compact = result.compact_string(include_species=True)
        # Compact string should not have * or : in allele portion
        assert ":" not in compact.split("*")[-1] if "*" in compact else True

    @pytest.mark.parametrize("allele_str,gene,fields,annotations", STANDARD_CLASS1_ALLELES)
    def test_to_record(self, allele_str, gene, fields, annotations):
        """Test dictionary record output."""
        result = parse(allele_str)
        record = result.to_record()
        assert "gene" in record
        assert "allele" in record
        assert "species_prefix" in record
        # gene field includes species prefix (e.g., "HLA-A" not just "A")
        assert gene in record["gene"]


# ============================================================================
# Four-field (8-digit) alleles - full resolution typing
# ============================================================================


class TestFourFieldAlleles:
    """Tests for full 4-field allele designations."""

    FOUR_FIELD_ALLELES = [
        "HLA-A*02:01:01:01",
        "HLA-A*01:01:01:01",
        "HLA-B*07:02:01:01",
        "HLA-B*08:01:01:01",
        "HLA-B*15:01:01:01",
        "HLA-C*07:02:01:01",
        "HLA-DRB1*01:01:01:01",
        "HLA-DRB1*03:01:01:01",
        "HLA-DRB1*04:01:01:01",
        "HLA-DQB1*02:01:01:01",
    ]

    @pytest.mark.parametrize("allele_str", FOUR_FIELD_ALLELES)
    def test_parse_four_field(self, allele_str):
        """Parse 4-field alleles."""
        result = parse(allele_str)
        assert result is not None
        assert type(result) is Allele
        eq_(len(result.allele_fields), 4)

    @pytest.mark.parametrize("allele_str", FOUR_FIELD_ALLELES)
    def test_restrict_allele_fields(self, allele_str):
        """Test restricting allele fields to lower resolution."""
        result = parse(allele_str)

        # Restrict to 2 fields
        two_field = result.restrict_allele_fields(2)
        eq_(len(two_field.allele_fields), 2)
        eq_(two_field.allele_fields, result.allele_fields[:2])

        # Restrict to 1 field (allele group only)
        one_field = result.restrict_allele_fields(1)
        eq_(len(one_field.allele_fields), 1)


# ============================================================================
# Alleles with expression annotations (N, L, S, C, A, Q)
# ============================================================================


class TestAnnotatedAlleles:
    """Tests for alleles with functional annotations."""

    ANNOTATED_ALLELES = [
        # Null alleles (not expressed)
        ("HLA-A*02:15N", "A", ("02", "15"), ("N",), "annotation_null"),
        ("HLA-B*15:01N", "B", ("15", "01"), ("N",), "annotation_null"),
        ("HLA-B*44:02:01:02N", "B", ("44", "02", "01", "02"), ("N",), "annotation_null"),
        # Low expression
        ("HLA-A*24:02L", "A", ("24", "02"), ("L",), "annotation_low_expression"),
        ("HLA-B*44:02L", "B", ("44", "02"), ("L",), "annotation_low_expression"),
        # Secreted (soluble)
        ("HLA-A*24:02S", "A", ("24", "02"), ("S",), "annotation_secreted"),
        # Cytoplasmic (not on cell surface)
        ("HLA-B*44:02C", "B", ("44", "02"), ("C",), "annotation_cystosolic"),
        # Questionable
        ("HLA-A*02:01Q", "A", ("02", "01"), ("Q",), "annotation_questionable"),
        # Aberrant expression
        ("HLA-A*02:01A", "A", ("02", "01"), ("A",), "annotation_aberrant_expression"),
    ]

    @pytest.mark.parametrize("allele_str,gene,fields,annotations,attr", ANNOTATED_ALLELES)
    def test_parse_annotated_allele(self, allele_str, gene, fields, annotations, attr):
        """Parse alleles with expression annotations."""
        result = parse(allele_str)
        assert result is not None, f"Failed to parse {allele_str}"
        assert type(result) is Allele
        eq_(result.gene_name, gene)
        eq_(result.allele_fields, fields)
        eq_(result.annotations, annotations)
        # Check the annotation property
        assert getattr(result, attr) is True


# ============================================================================
# HLA Class II alleles - alpha/beta chains
# ============================================================================


class TestClassIIAlleles:
    """Tests for HLA Class II allele formats."""

    CLASS2_BETA_ALLELES = [
        # DR locus
        ("HLA-DRB1*01:01", "DRB1", ("01", "01")),
        ("HLA-DRB1*03:01", "DRB1", ("03", "01")),
        ("HLA-DRB1*04:01", "DRB1", ("04", "01")),
        ("HLA-DRB1*04:04", "DRB1", ("04", "04")),
        ("HLA-DRB1*07:01", "DRB1", ("07", "01")),
        ("HLA-DRB1*11:01", "DRB1", ("11", "01")),
        ("HLA-DRB1*13:01", "DRB1", ("13", "01")),
        ("HLA-DRB1*15:01", "DRB1", ("15", "01")),  # MS association
        ("HLA-DRB3*01:01", "DRB3", ("01", "01")),
        ("HLA-DRB4*01:01", "DRB4", ("01", "01")),
        ("HLA-DRB5*01:01", "DRB5", ("01", "01")),
        # DQ locus
        ("HLA-DQB1*02:01", "DQB1", ("02", "01")),
        ("HLA-DQB1*03:01", "DQB1", ("03", "01")),
        ("HLA-DQB1*03:02", "DQB1", ("03", "02")),  # T1D association
        ("HLA-DQB1*05:01", "DQB1", ("05", "01")),
        ("HLA-DQB1*06:02", "DQB1", ("06", "02")),
        # DP locus
        ("HLA-DPB1*01:01", "DPB1", ("01", "01")),
        ("HLA-DPB1*02:01", "DPB1", ("02", "01")),
        ("HLA-DPB1*04:01", "DPB1", ("04", "01")),
        ("HLA-DPB1*04:02", "DPB1", ("04", "02")),
    ]

    CLASS2_ALPHA_ALLELES = [
        ("HLA-DRA*01:01", "DRA", ("01", "01")),
        ("HLA-DRA*01:02", "DRA", ("01", "02")),
        ("HLA-DQA1*01:01", "DQA1", ("01", "01")),
        ("HLA-DQA1*01:02", "DQA1", ("01", "02")),
        ("HLA-DQA1*03:01", "DQA1", ("03", "01")),
        ("HLA-DQA1*05:01", "DQA1", ("05", "01")),
        ("HLA-DPA1*01:03", "DPA1", ("01", "03")),
        ("HLA-DPA1*02:01", "DPA1", ("02", "01")),
    ]

    @pytest.mark.parametrize("allele_str,gene,fields", CLASS2_BETA_ALLELES)
    def test_parse_class2_beta(self, allele_str, gene, fields):
        """Parse Class II beta chain alleles."""
        result = parse(allele_str)
        assert result is not None
        assert type(result) is Allele
        eq_(result.gene_name, gene)
        eq_(result.allele_fields, fields)
        assert result.is_class2
        assert result.is_class2_beta

    @pytest.mark.parametrize("allele_str,gene,fields", CLASS2_ALPHA_ALLELES)
    def test_parse_class2_alpha(self, allele_str, gene, fields):
        """Parse Class II alpha chain alleles."""
        result = parse(allele_str)
        assert result is not None
        assert type(result) is Allele
        eq_(result.gene_name, gene)
        eq_(result.allele_fields, fields)
        assert result.is_class2
        assert result.is_class2_alpha


# ============================================================================
# Class II alpha/beta pairs
# ============================================================================


class TestClassIIPairs:
    """Tests for Class II alpha/beta pair formats."""

    CLASS2_PAIRS = [
        # DR pairs
        ("HLA-DRA*01:01/DRB1*01:01", "DRA", "01:01", "DRB1", "01:01"),
        ("HLA-DRA*01:01/DRB1*03:01", "DRA", "01:01", "DRB1", "03:01"),
        ("HLA-DRA*01:01/DRB1*04:01", "DRA", "01:01", "DRB1", "04:01"),
        ("HLA-DRA*01:01/DRB1*15:01", "DRA", "01:01", "DRB1", "15:01"),
        # DQ pairs
        ("HLA-DQA1*01:02/DQB1*06:02", "DQA1", "01:02", "DQB1", "06:02"),
        ("HLA-DQA1*05:01/DQB1*02:01", "DQA1", "05:01", "DQB1", "02:01"),
        ("HLA-DQA1*03:01/DQB1*03:02", "DQA1", "03:01", "DQB1", "03:02"),
        # DP pairs
        ("HLA-DPA1*01:03/DPB1*04:01", "DPA1", "01:03", "DPB1", "04:01"),
        ("HLA-DPA1*02:01/DPB1*01:01", "DPA1", "02:01", "DPB1", "01:01"),
    ]

    @pytest.mark.parametrize("pair_str,alpha_gene,alpha_fields,beta_gene,beta_fields", CLASS2_PAIRS)
    def test_parse_class2_pair(self, pair_str, alpha_gene, alpha_fields, beta_gene, beta_fields):
        """Parse Class II alpha/beta pairs."""
        result = parse(pair_str)
        assert result is not None, f"Failed to parse {pair_str}"
        assert type(result) is Pair

        # Check alpha chain
        assert result.alpha is not None
        eq_(result.alpha.gene_name, alpha_gene)

        # Check beta chain
        assert result.beta is not None
        eq_(result.beta.gene_name, beta_gene)

    def test_pair_to_string_roundtrip(self):
        """Test that pair can be converted to string and reparsed."""
        pair_str = "HLA-DRA*01:01/DRB1*01:01"
        result = parse(pair_str)
        output = result.to_string()
        reparsed = parse(output)
        eq_(type(reparsed), Pair)


# ============================================================================
# Alternate input formats commonly seen in the wild
# ============================================================================


class TestAlternateFormats:
    """Tests for non-standard but common allele formats."""

    ALTERNATE_FORMATS = [
        # Without HLA prefix
        ("A*02:01", "A", ("02", "01")),
        ("B*07:02", "B", ("07", "02")),
        ("DRB1*01:01", "DRB1", ("01", "01")),
        # Lowercase
        ("hla-a*02:01", "A", ("02", "01")),
        ("hla-b*07:02", "B", ("07", "02")),
        # Without colons (compact format)
        ("HLA-A*0201", "A", ("02", "01")),
        ("HLA-B*0702", "B", ("07", "02")),
        ("A*0201", "A", ("02", "01")),
        # With underscore separator
        ("HLA-A_0201", "A", ("02", "01")),
        ("A_02_01", "A", ("02", "01")),
        # Ultra-compact (no separators)
        ("A0201", "A", ("02", "01")),
        ("B0702", "B", ("07", "02")),
        # Three-digit allele groups (high-resolution typing)
        ("HLA-B*15:120", "B", ("15", "120")),
        ("HLA-A*02:101", "A", ("02", "101")),
    ]

    @pytest.mark.parametrize("allele_str,gene,fields", ALTERNATE_FORMATS)
    def test_parse_alternate_format(self, allele_str, gene, fields):
        """Parse alternate allele formats."""
        result = parse(allele_str)
        assert result is not None, f"Failed to parse {allele_str}"
        assert type(result) is Allele
        eq_(result.gene_name, gene)
        eq_(result.allele_fields, fields)


# ============================================================================
# Serotypes (serological typing)
# ============================================================================


class TestSerotypes:
    """Tests for serological type designations."""

    SEROTYPES = [
        ("HLA-A2", "A2"),
        ("HLA-A1", "A1"),
        ("HLA-A3", "A3"),
        ("HLA-B7", "B7"),
        ("HLA-B8", "B8"),
        ("HLA-B27", "B27"),
        ("HLA-B35", "B35"),
        ("HLA-DR1", "DR1"),
        ("HLA-DR3", "DR3"),
        ("HLA-DR4", "DR4"),
        ("HLA-DR7", "DR7"),
        ("HLA-DQ2", "DQ2"),
    ]

    @pytest.mark.parametrize("sero_str,name", SEROTYPES)
    def test_parse_serotype(self, sero_str, name):
        """Parse serological types."""
        result = parse(sero_str)
        assert result is not None, f"Failed to parse {sero_str}"
        assert type(result) is Serotype
        eq_(result.name, name)
        assert result.species.is_human


# ============================================================================
# Mutations
# ============================================================================


class TestMutantAlleles:
    """Tests for alleles with point mutations."""

    def test_single_mutation(self):
        """Parse allele with single mutation."""
        result = parse("HLA-A*02:01 T80M mutant")
        assert result is not None
        assert result.is_mutant
        eq_(len(result.mutations), 1)
        eq_(result.mutations[0].aa_original, "T")
        eq_(result.mutations[0].pos, 80)
        eq_(result.mutations[0].aa_mutant, "M")

    def test_multiple_mutations(self):
        """Parse allele with multiple mutations."""
        result = parse("HLA-A*02:01 E152A, R155Y, L156Y mutant")
        assert result is not None
        assert result.is_mutant
        eq_(len(result.mutations), 3)

    def test_class2_pair_with_mutation(self):
        """Parse Class II pair with mutation on beta chain."""
        result = parse("HLA-DRA*01:01/DRB1*01:01 G86Y mutant")
        assert result is not None
        assert type(result) is Pair
        # Mutation should be on beta chain
        assert result.beta.is_mutant
        eq_(len(result.beta.mutations), 1)


# ============================================================================
# Non-classical HLA genes
# ============================================================================


class TestNonClassicalHLA:
    """Tests for non-classical HLA molecules."""

    NON_CLASSICAL_ALLELES = [
        # HLA-E
        ("HLA-E*01:01", "E", ("01", "01")),
        ("HLA-E*01:03", "E", ("01", "03")),
        # HLA-F
        ("HLA-F*01:01", "F", ("01", "01")),
        # HLA-G
        ("HLA-G*01:01", "G", ("01", "01")),
        ("HLA-G*01:04", "G", ("01", "04")),
        # MICA/MICB
        ("HLA-MICA*001", "MICA", ("001",)),
        ("HLA-MICA*002", "MICA", ("002",)),
        ("HLA-MICA*008", "MICA", ("008",)),
        ("HLA-MICB*002", "MICB", ("002",)),
    ]

    @pytest.mark.parametrize("allele_str,gene,fields", NON_CLASSICAL_ALLELES)
    def test_parse_non_classical(self, allele_str, gene, fields):
        """Parse non-classical HLA alleles."""
        result = parse(allele_str)
        assert result is not None, f"Failed to parse {allele_str}"
        assert type(result) is Allele
        eq_(result.gene_name, gene)
        eq_(result.allele_fields, fields)


# ============================================================================
# Gene-level parsing (without allele designation)
# ============================================================================


class TestGeneParsing:
    """Tests for gene-level parsing."""

    GENES = [
        ("HLA-A", "A"),
        ("HLA-B", "B"),
        ("HLA-C", "C"),
        ("HLA-DRA", "DRA"),
        ("HLA-DRB1", "DRB1"),
        ("HLA-DQA1", "DQA1"),
        ("HLA-DQB1", "DQB1"),
        ("HLA-DPA1", "DPA1"),
        ("HLA-DPB1", "DPB1"),
    ]

    @pytest.mark.parametrize("gene_str,name", GENES)
    def test_parse_gene(self, gene_str, name):
        """Parse gene names."""
        result = parse(gene_str)
        assert result is not None, f"Failed to parse {gene_str}"
        assert type(result) is Gene
        eq_(result.name, name)


# ============================================================================
# MHC Class parsing
# ============================================================================


class TestMhcClassParsing:
    """Tests for MHC class-level parsing."""

    def test_parse_hla_class_i(self):
        """Parse HLA Class I."""
        result = parse("HLA class I")
        assert result is not None
        assert type(result) is MhcClass
        assert result.is_class1

    def test_parse_hla_class_ii(self):
        """Parse HLA Class II."""
        result = parse("HLA class II")
        assert result is not None
        assert type(result) is MhcClass
        assert result.is_class2


# ============================================================================
# Species parsing
# ============================================================================


class TestSpeciesParsing:
    """Tests for species-level parsing."""

    def test_parse_hla_species(self):
        """Parse HLA as species."""
        result = parse("HLA")
        assert result is not None
        assert type(result) is Species
        assert result.is_human

    def test_species_from_allele(self):
        """Verify species is correctly identified from allele."""
        result = parse("HLA-A*02:01")
        assert result.species.is_human
        eq_(result.species.prefix, "HLA")


# ============================================================================
# Error handling
# ============================================================================


class TestErrorHandling:
    """Tests for error handling."""

    INVALID_INPUTS = [
        "",  # Empty string
        "   ",  # Whitespace only
        "INVALID*99:99",  # Invalid gene
        "HLA-Z*01:01",  # Non-existent gene
    ]

    @pytest.mark.parametrize("invalid_str", INVALID_INPUTS)
    def test_parse_invalid_raises(self, invalid_str):
        """Invalid inputs should raise ParseError."""
        with pytest.raises(ParseError):
            parse(invalid_str)

    @pytest.mark.parametrize("invalid_str", INVALID_INPUTS)
    def test_parse_invalid_returns_none(self, invalid_str):
        """Invalid inputs should return None with raise_on_error=False."""
        result = parse(invalid_str, raise_on_error=False)
        assert result is None


# ============================================================================
# Edge cases and boundary conditions
# ============================================================================


class TestEdgeCases:
    """Tests for edge cases and boundary conditions."""

    def test_allele_with_leading_zeros(self):
        """Alleles with leading zeros should be preserved."""
        result = parse("HLA-A*01:01")
        eq_(result.allele_fields, ("01", "01"))
        # Verify leading zeros preserved in output
        assert "01:01" in result.to_string()

    def test_very_long_allele_name(self):
        """Parse 4-field allele with all fields."""
        result = parse("HLA-A*02:01:01:01")
        eq_(len(result.allele_fields), 4)

    def test_allele_equality(self):
        """Two alleles parsed from same string should be equal."""
        a1 = parse("HLA-A*02:01")
        a2 = parse("HLA-A*02:01")
        eq_(a1, a2)
        assert hash(a1) == hash(a2)

    def test_allele_inequality_different_fields(self):
        """Alleles with different fields should not be equal."""
        a1 = parse("HLA-A*02:01")
        a2 = parse("HLA-A*02:02")
        assert a1 != a2

    def test_allele_inequality_different_genes(self):
        """Alleles from different genes should not be equal."""
        a1 = parse("HLA-A*02:01")
        a2 = parse("HLA-B*02:01")
        assert a1 != a2

    def test_case_insensitivity(self):
        """Parsing should be case insensitive."""
        upper = parse("HLA-A*02:01")
        lower = parse("hla-a*02:01")
        mixed = parse("Hla-A*02:01")
        eq_(upper, lower)
        eq_(upper, mixed)


# ============================================================================
# Real-world clinical/research scenarios
# ============================================================================


class TestRealWorldScenarios:
    """Tests simulating real-world usage scenarios."""

    def test_clinical_typing_report(self):
        """Simulate parsing a clinical HLA typing report."""
        typing_results = [
            "HLA-A*02:01",
            "HLA-A*03:01",
            "HLA-B*07:02",
            "HLA-B*44:02",
            "HLA-C*05:01",
            "HLA-C*07:02",
            "HLA-DRB1*03:01",
            "HLA-DRB1*07:01",
            "HLA-DQB1*02:01",
            "HLA-DQB1*03:03",
        ]

        for allele_str in typing_results:
            result = parse(allele_str)
            assert result is not None
            assert type(result) is Allele
            assert result.species.is_human

    def test_netmhcpan_output_format(self):
        """Parse alleles in NetMHCpan output format."""
        netmhcpan_alleles = [
            "HLA-A02:01",  # NetMHCpan sometimes omits the *
            "HLA-B07:02",
            "HLA-B57:01",
        ]

        for allele_str in netmhcpan_alleles:
            result = parse(allele_str)
            assert result is not None

    def test_iedb_format(self):
        """Parse alleles in IEDB format."""
        iedb_alleles = [
            "HLA-A*02:01",
            "HLA-A*0201",  # IEDB sometimes uses compact format
            "H-2-Kb",  # Mouse
            "H-2-Db",
        ]

        for allele_str in iedb_alleles:
            result = parse(allele_str)
            assert result is not None

    def test_disease_association_alleles(self):
        """Parse alleles commonly associated with diseases."""
        disease_alleles = {
            "HLA-B*27:05": "Ankylosing spondylitis",
            "HLA-DRB1*15:01": "Multiple sclerosis",
            "HLA-DQB1*03:02": "Type 1 diabetes",
            "HLA-C*06:02": "Psoriasis",
            "HLA-B*57:01": "HIV control / Abacavir hypersensitivity",
            "HLA-B*58:01": "Allopurinol hypersensitivity",
            "HLA-DRB1*04:01": "Rheumatoid arthritis",
        }

        for allele_str, disease in disease_alleles.items():
            result = parse(allele_str)
            assert result is not None, f"Failed to parse {allele_str} ({disease})"
            assert type(result) is Allele



from mhcgnomes import Pair, parse

from .common import eq_


def test_restrict_num_allele_fields_HLA_DRA_01_01_01_01_DRB1_01_01_01_01():
    result = parse("DRA*01:01:01:01/DRB*01:01:01:01")
    eq_(type(result), Pair)
    eq_(result.alpha.allele_fields, ("01", "01", "01", "01"))
    eq_(result.beta.allele_fields, ("01", "01", "01", "01"))

    result2 = result.restrict_allele_fields(2)
    eq_(
        result2.alpha.allele_fields,
        (
            "01",
            "01",
        ),
    )
    eq_(
        result2.beta.allele_fields,
        (
            "01",
            "01",
        ),
    )


def test_annotation_null_HLA_DRA_01_01_01_01_DRB1_01_01_01_01N():
    result = parse("DRA*01:01:01:01/DRB*01:01:01:01")
    eq_(type(result), Pair)
    assert not result.annotation_null

    result = parse("DRA*01:01:01:01/DRB*01:01:01:01N")
    eq_(type(result), Pair)
    assert result.annotation_null



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

from .errors import ParseError
from .mhc_class_helpers import (
    is_class1,
    is_class2,
    is_valid_restriction,
    normalize_mhc_class_string,
)
from .result_with_mhc_class import ResultWithMhcClass
from .species import Species


class MhcClass(ResultWithMhcClass):
    """
    Wrapper class for species combined with MHC classes such as
    "I", "Ia", "Ib", "II", "IIa", &c
    which provides some utility functions.
    """

    def __init__(self, species: Species, mhc_class: str, raw_string: Union[str, None] = None):
        ResultWithMhcClass.__init__(
            self, species=species, mhc_class=mhc_class, raw_string=raw_string
        )

    @classmethod
    def get(cls, species_prefix, mhc_class):
        species = Species.get(species_prefix)
        if species is None:
            return None
        try:
            mhc_class = normalize_mhc_class_string(mhc_class)
        except ParseError:
            return None
        return MhcClass(species, mhc_class)

    @property
    def is_class1(self):
        return is_class1(self.mhc_class)

    @property
    def is_class2(self):
        return is_class2(self.mhc_class)

    def genes(self):
        """
        Returns all gene names whose MHC class matches this MHC class
        """
        return [
            g
            for g in self.species.genes()
            if is_valid_restriction(self.mhc_class, self.get_mhc_class_of_gene(g))
        ]

    def to_record(self):
        d = self.species.to_record()
        d["mhc_class"] = self.mhc_class
        return d

    def to_string(self, include_species=True, use_old_species_prefix=False):
        if include_species:
            if self.common_species_name:
                species_str = self.common_species_name.lower()
            elif use_old_species_prefix:
                species_str = self.species.historic_alias
            else:
                species_str = self.species.prefix
            return f"{species_str} class {self.mhc_class}"
        else:
            return f"class {self.mhc_class}"

    def compact_string(self, include_species=True, use_old_species_prefix=False):
        """
        Compact representation of an MHC class, currently same as the
        normalized representation.
        """
        return self.to_string(
            include_species=include_species, use_old_species_prefix=use_old_species_prefix
        )
